//
//  IKAROS.cc		Kernel code for the IKAROS project
//			        Version 1.0.0
//
//    Copyright (C) 2001-2007  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    See http://www.ikaros-project.org/ for more information.
//
//	Created July 13, 2001
//
//	Revision History:
//
//		2002-01-27	Added support for sizing of output arrays depending on input sizes
//		2002-02-06	Added x, y of each array instead of width
//		2002-02-10	Added new XML support that does not depend on expat for better portability
//		2002-02-16	Additional support for matrices: GetInputMatrix() and GetOutputMatrix()
//					to make image processing easier
//		2002-03-15	Added socket for UI communication
//		2002-05-20 	Added new communication protocol
//		2002-10-06	Minor changes in parameter class
//		2003-01-23	Bug in InitInputs() fixed
//		2003-08-09	Now catches ctrl-C to shut down gracefully
//
//		2004-03-02	Version 0.8.0 created
//		2004-11-15	New print and error functions; now use Notify(msg, ...) for both errors and printing
//		2004-11-27	Added defines around CTRL-C handler
//		2005-01-17	Changed Data to Array
//		2005-01-18	Completed timing functions
//		2005-08-31	All communication code moved to WebUI
//		2006-01-20	Major cleanup of the code; most system specific code moved out of the kernel
//		2006-02-10	Extended error handling
//		2006-05-05	Even more extended error handling; more informative error messages
//		2006-08-31	Most Windows-specific code included
//		2006-12-12	Fixed potential memory leaks caused by old XML parser
//		2007-01-10	Added new XML handling and new module creation for ikc files
//		2007-05-10	Version 1.0.0 created


#include "IKAROS.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>
#include <fcntl.h>
#include <ctype.h>

#ifdef WINDOWS
#include <direct.h>
#else
#include <unistd.h>
#endif

#ifdef WINDOWS32
#include <windows.h>
#undef GetClassName
#else
#include <unistd.h>
#endif

#ifdef USE_SAFE_NEW
#include <exception> // for std::bad_alloc
#include <new>
#include <cstdlib> // for malloc() and free()
#endif



using namespace ikaros;
bool			global_fatal_error = false;	// Must be global because it is used before the kernel is created
bool			global_terminate = false;	// Used to flag that CTRL-C has been received
Parameter *     global_parameter = NULL;	// Used to pass parameters to modules on creation for backward compatibility; will be removed in future versions


//
// USE_SAFE_NEW redefines new to automatically clear memory.
// This avoids many bugs when all pointers have not been cleared
// in a module when an error casues the initialization process
// to stop.
//

#ifdef USE_SAFE_NEW

void* operator new (std::size_t size) throw (std::bad_alloc)
{
    void *p=calloc(size, sizeof(char)); 
    if (p==0) // did calloc succeed?
        throw std::bad_alloc();
    return p;
}

void operator delete (void *p) throw()
{
    free(p); 
}

#endif

//
// Class
//
Class::Class(char * n, ModuleCreator mc, char * p, Class * nxt)
{
    name = n;
    module_creator = mc;
    path = p;
    next = nxt;
}

Class::~Class()
{
// path
    delete next;
}

char *
Class::GetClassPath(char * class_name)
{
    if (equal_strings(name, class_name))
        return path;
    else if (next != NULL)
        return next->GetClassPath(class_name);
    else
        return NULL;
}

Module *
CreateModule(Class * c, char * class_name, char * module_name, Parameter * p)
{
    if (c == NULL)
    {
        return NULL;
    }
    else if (equal_strings(c->name, class_name))
    {
        global_parameter = p; // Pass values also in a global for backward compatibility
        return (*c->module_creator)(module_name, p);
    }
    else
        return CreateModule(c->next, class_name, module_name, p);
}

void
Module_IO::Allocate()
{
    if (sizex == unknown_size || sizey == unknown_size)
    {
        if (module != NULL)
            module->Notify(msg_fatal_error, "Attempting to allocate io (\"%s\") with unknown size for module \"%s\" (%s). Check that all required inputs are connected.\n", name, module->GetName(), module->GetClassName());
        return;
    }
    if (module != NULL) module->Notify(msg_verbose, "Allocating data of size %d.\n", size);
    data	=   new float * [max_delay];
    matrix  =   new float ** [max_delay];
    for (int d=0; d<max_delay; d++)
    {
        matrix[d] = create_matrix(sizex, sizey);
        data[d] = matrix[d][0];
    }
}

Module_IO::Module_IO(Module_IO * nxt, Module * m, char * n, int x, int y)
{
    next = nxt;
    module = m;
    name = n;
    data = NULL;
    matrix = NULL;
    size = x*y;
    sizex = x;
    sizey = y;
    max_delay = 1;
}

Module_IO::~Module_IO()
{
    if (name != NULL)
    {
        if (module != NULL) module->Notify(msg_verbose, "      Deleting Module_IO \"%s\".\n", name);
    }
    else
    {
        if (module != NULL) module->Notify(msg_verbose, "      Deleting Module_IO\n");
    }
    if (matrix)
        for (int d=0; d<max_delay; d++)
            destroy_matrix(matrix[d]);
    delete [] data;
    delete [] matrix;
    delete next;
}

void
Module_IO::SetSize(int x, int y)
{
    int s = x*y;
    if (x == unknown_size)
        return;
    if (size != unknown_size && s != size)
    {
        if (module != NULL)
            module->Notify(msg_warning, "Module_IO::SetSize: Attempt to resize data array \"%s\" of module \"%s\" (%s) (%d <= %d). Ignored.\n",  name, module->GetName(), module->GetClassName(), size, s);
        return;
    }
    if (s == size)
        return;
    if (s == 0)
        return;
    if (module != NULL) module->Notify(msg_verbose, "Allocating memory for input/output \"%s\" of module \"%s\" (%s) with size %d and max_delay = %d (in SetSize).\n", name, module->name, module->GetClassName(), s, max_delay);
    sizex = x;
    sizey = y;
    size = x*y;
    if (module != NULL && module->kernel != NULL)
        module->kernel->NotifySizeChange();
}

void
Module_IO::DelayOutputs()
{
    for (int d=max_delay-1; d>0; d--)
        copy_matrix(matrix[d], matrix[d-1], sizex, sizey);
}

Module::~Module()
{
    Notify(msg_verbose, "    Deleting Module \"%s\".\n", name);
    delete input_list;
    delete output_list;
    delete next;
}

void
Module::AddInput(char * name)
{
    if (GetModule_IO(input_list, name) != NULL)
    {
        Notify(msg_warning, "Input \"%s\" of module \"%s\" (%s) already exists.\n", name, GetName(), GetClassName());
        return;
    }
    input_list = new Module_IO(input_list, this, name);
    Notify(msg_verbose, "  Adding input \"%s\".\n", name);
}

void
Module::AddOutput(char * name, int sizeX, int sizeY)
{
    if (GetModule_IO(output_list, name) != NULL)
    {
        Notify(msg_warning, "Output \"%s\" of module \"%s\" (%s) already exists.\n", name, GetName(), GetClassName());
        return;
    }
    output_list = new Module_IO(output_list, this, name, sizeX, sizeY);
    Notify(msg_verbose, "  Adding output \"%s\" of size %d x %d to module \"%s\" (%s).\n", name, sizeX, sizeY, GetName(), GetClassName());
}

char *
Module::GetName()
{
    return name;
}

char *
Module::GetClassName()
{
    return class_name;
}

char *
Module::GetValue(char * n)	// This function implements attribute inheritance with renaming through the parameter element
{
    char * module_name = GetName();
    // Check for local value in this element
    char * value = xml->GetAttribute(n);
    if (value != NULL)
        return value;
    // not found here, loop up the group hierarchy
    for (XMLElement * parent = xml->GetParentElement(); parent != NULL; parent = parent->GetParentElement())
    {
        // Look for parameter element that redefines the attribute name
        for (XMLElement * parameter = parent->GetElement("parameter"); parameter != NULL; parameter = parameter->next->FindElement("parameter"))
        {
            char * t = parameter->GetAttribute("target");
            if (equal_strings(parameter->GetAttribute("target"), n))
            {
                // we have found our parameter
                // it controls this module if module name is set to the name of this module or if it is not set = relates to all modules
                char * tm = parameter->GetAttribute("module");
                if (tm == NULL || (equal_strings(tm, module_name)))
                {
                    // the parameter element redefines our parameter name; get the new name
                    char * newname = parameter->GetAttribute("name");
                    if (newname == NULL)
                    {
                        Notify(msg_fatal_error, "A parameter element with target \"%s\" lacks a name attribute.\n", t);
                        return NULL;
                    }
                    // we have the new name; set it and see if it is defined in the current group element
                    n = newname;
                }
            }
        }
        value = parent->GetAttribute(n);
        if (value != NULL)
            return value;
        // It was not found here; shift module name to that of the current group and continue up the group hierarchy...
        module_name = parent->GetAttribute("name");
    }
    return NULL;
}

float
Module::GetFloatValue(char * n, float d)
{
    return string_to_float(GetValue(n), d);
}

int
Module::GetIntValue(char * n, int d)
{
    return string_to_int(GetValue(n), d);
}

static bool
tobool(char * v)
{
    if (!v) return false;
    if (!strcmp(v, "true")) return true;
    if (!strcmp(v, "True")) return true;
    if (!strcmp(v, "TRUE")) return true;
    if (!strcmp(v, "yes")) return true;
    if (!strcmp(v, "Yes")) return true;
    if (!strcmp(v, "YES")) return true;
    if (!strcmp(v, "1")) return true;
    if (!strcmp(v, "false")) return false;
    if (!strcmp(v, "False")) return false;
    if (!strcmp(v, "FALSE")) return false;
    if (!strcmp(v, "no")) return false;
    if (!strcmp(v, "No")) return false;
    if (!strcmp(v, "NO")) return false;
    if (!strcmp(v, "0")) return false;
    return false;
}

bool
Module::GetBoolValue(char * n, bool d)
{
    char * v = GetValue(n);
    if (v == NULL)
        return d;
    else
        return tobool(v);
}

static int
findindex(char * name, char * list)
{
    int ix = 0;
    int lp = 0;
    while (true)
    {
        // try to match name in list
        int np = 0;
        while (list[lp] && list[lp] != '/'  && name[np] == list[lp])
        {
            np++;
            lp++;
        }
        // name matches, return index
        if ((list[lp] == '/' || list[lp] == 0) && name[np] == 0)
            return ix;
        // skip to next name
        ix++;
        while (list[lp] && list[lp] != '/')
            lp++;
        // no match
        if (list[lp] == 0)
            return 0;
        // prepare for next match
        lp++;
    }
}

int
Module::GetIntValueFromList(char * n, char * list)
{
    char * v = GetValue(n);
    if (v == NULL)
        return 0;
    else
        return findindex(v, list);
}

float *
Module::GetArray(char *n, int size)
{
    float * a = create_array(size);
    char * v = GetValue(n);
    if (v == NULL)
        return a;
    for (int i=0; i<size;i++)
    {
        for (; isspace(*v) && *v != '\0'; v++) ;
        if (sscanf(v, "%f", &a[i])==-1)
        {
            Notify(msg_warning, "Too few constants in matrix(0 assumed).\n");
            a[i] = 0;
        }
        for (; !isspace(*v) && *v != '\0'; v++) ;
    }
    return a;
}

float **
Module::GetMatrix(char *n, int sizex, int sizey)
{
    float ** m = create_matrix(sizex, sizey);
    float * d = m[0];
    char * v = GetValue(n);
    if (v == NULL)
        return m;
    for (int i=0; i<sizex*sizey;i++)
    {
        for (; isspace(*v) && *v != '\0'; v++) ;
        if (sscanf(v, "%f", &d[i])==-1)
        {
            Notify(msg_warning, "Too few constants in matrix(0 assumed).\n");
            d[i] = 0;
        }
        for (; !isspace(*v) && *v != '\0'; v++) ;
    }
    return m;
}

Module_IO *
Module::GetModule_IO(Module_IO * list, const char * name)
{
    for (Module_IO * i = list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
            return i;
    return NULL;
}

void
Module::AllocateOutputs()
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        i->Allocate();
}

void
Module::DelayOutputs()
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        i->DelayOutputs();
}

float *
Module::GetInputArray(char * name, bool required)
{
    for (Module_IO * i = input_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
        {
            if (i->data == NULL)
            {
                if(required)
                    Notify(msg_fatal_error, "Input array \"%s\" of module \"%s\" (%s) has no allocated data. Returning NULL.\n", name, GetName(), GetClassName());
                return NULL;
            }
            else
                return i->data[0];
        }
    Notify(msg_warning, "Input array \"%s\" of module \"%s\" (%s) does not exist. Returning NULL.\n", name, GetName(), GetClassName());
    return NULL;
}

float *
Module::GetOutputArray(char * name, bool required)
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
        {
            if (i->data == NULL)
            {
                if(required)
                    Notify(msg_fatal_error, "Output array \"%s\" of module \"%s\" (%s) has no allocated data. Returning NULL.\n", name, GetName(), GetClassName());
                return NULL;
            }
            else
                return i->data[0];
        }
    Notify(msg_warning, "Output array  \"%s\" of module \"%s\" (%s) does not exist. Returning NULL.\n", name, GetName(), GetClassName());
    return NULL;
}

float **
Module::GetInputMatrix(char * name, bool required)
{
    for (Module_IO * i = input_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
        {
            if (i->matrix == NULL)
            {
                if(required)
                    Notify(msg_fatal_error, "Input matrix \"%s\" of module \"%s\" (%s) has no allocated data. Returning NULL.\n", name, GetName(), GetClassName());
                return NULL;
            }
            else
                return i->matrix[0];
        }
    Notify(msg_warning, "Input matrix \"%s\" of module \"%s\" (%s) does not exist. Returning NULL.\n", name, GetName(), GetClassName());
    return NULL;
}

float **
Module::GetOutputMatrix(char * name, bool required)
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
        {
            if (i->matrix == NULL)
            {
                if(required)
                    Notify(msg_fatal_error, "Output matrix \"%s\" of module \"%s\" (%s) has no allocated data. Returning NULL.\n", name, GetName(), GetClassName());
                return NULL;
            }
            else
                return i->matrix[0];
        }
    Notify(msg_warning, "Output matrix \"%s\" of module \"%s\" (%s) does not exist. Returning NULL.\n", name, GetName(), GetClassName());
    return NULL;
}

int
Module::GetInputSize(char * input_name)
{
    // Find the Module_IO for this input
    for (Module_IO * i = input_list; i != NULL; i = i->next)
        if (equal_strings(input_name, i->name))
        {
            if (i->size != unknown_size)
                return i->size;
            else if (kernel != NULL)
                return kernel->CalculateInputSize(i);
            else
                break;
        }
    Notify(msg_fatal_error, "Cannot calculate input size for \"%s\" of module \"%s\" (%s).\n", input_name, name, class_name);
    return unknown_size;
}

int
Module::GetInputSizeX(char * input_name)
{
    // Find the Module_IO for this input
    for (Module_IO * i = input_list; i != NULL; i = i->next)
        if (equal_strings(input_name, i->name))
        {
            if (i->sizex != unknown_size)
                return i->sizex;
            else if (kernel != NULL)
                return kernel->CalculateInputSizeX(i);
            else
                break;
        }
    Notify(msg_fatal_error, "Cannot calculate input x size for \"%s\" of module \"%s\" (%s).\n", input_name, name, class_name);
    return unknown_size;
}

int
Module::GetInputSizeY(char * input_name)
{
    // Find the Module_IO for this input
    for (Module_IO * i = input_list; i != NULL; i = i->next)
        if (equal_strings(input_name, i->name))
        {
            if (i->sizex != unknown_size)	// Yes, sizeX is correct
                return i->sizey;
            else if (kernel != NULL)
                return kernel->CalculateInputSizeY(i);
            else
                break;
        }
    Notify(msg_fatal_error, "Cannot calculate input y size for \"%s\" of module \"%s\" (%s).\n", input_name, name, class_name);
    return unknown_size;
}

int
Module::GetOutputSize(char * name)
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
            return i->size;
    return 0;
}

int
Module::GetOutputSizeX(char * name)
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
            return  i->sizex;
    return 0;
}

int
Module::GetOutputSizeY(char * name)
{
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
            return  i->sizey;
    return 0;
}

bool
Module::InputConnected(char * name)
{
    return kernel->InputConnected(this, name);
}

bool
Module::OutputConnected(char * name)
{
    return kernel->OutputConnected(this, name);
}

void
Module::SetOutputSize(char * name, int x, int y)
{
    if (x < -1 || y < -1)
    {
        Notify(msg_warning, "Attempting to set negative size of  %s.%s \n", this->name, name);
        return;
    }
    for (Module_IO * i = output_list; i != NULL; i = i->next)
        if (equal_strings(name, i->name))
            i->SetSize(x, y);
}

// The contructor Module::Module(char * n) is deprectaed.
// Change to Module::Module(Parameter * p) instead.
Module::Module(char * n)
{
    Parameter * p = global_parameter;
    next = NULL;
    next_in_threadGroup = NULL;
    input_list = NULL;
    output_list = NULL;
    timer = new Timer();
    time = 0;
    ticks = 0;
    kernel = p->kernel;
    xml = p->xml;
    name = xml->GetAttribute("name");
    class_name = xml->GetAttribute("class");
    period = GetIntValue("period", 1);
    phase = GetIntValue("phase", 0);
    kernel->Notify(msg_warning, "Module \"%s\" (%s) uses DEPRECATED contructor Module::Module(char *).\n", name, class_name);
}

Module::Module(Parameter * p)
{
    next = NULL;
    next_in_threadGroup = NULL;
    input_list = NULL;
    output_list = NULL;
    timer = new Timer();
    time = 0;
    ticks = 0;
    kernel = p->kernel;
    xml = p->xml;
    name = GetValue("name");
    class_name = xml->GetAttribute("class");
    period = GetIntValue("period", 1);
    phase = GetIntValue("phase", 0);
}

void
Module::Notify(int msg)
{
    if (kernel != NULL)
        kernel->Notify(msg, "\n");
    else if (msg == msg_fatal_error)
        global_fatal_error = true;
}

void
Module::Notify(int msg, char *format, ...)
{
    char 	message[512];
    int n = 0;
    if (message == NULL)
    {
        global_fatal_error = true;
        return;
    }
    va_list 	args;
    va_start(args, format);
    vsnprintf(&message[n], 512, format, args);
    va_end(args);
    if (kernel != NULL)
    {
        kernel->Notify(msg, message);
    }
    else if (msg == msg_fatal_error)
    {
        global_fatal_error = true;
        printf("IKAROS: ERROR: %s", message);
    }
}

Connection::Connection(Connection * n, Module_IO * sio, int so, Module_IO * tio, int to, int s, int d)
{
    source_io = sio;
    source_offset = so;
    target_io = tio;
    target_offset = to;
    size = s;
    delay = d;
    next = n;
}

Connection::~Connection()
{
    if (source_io != NULL && source_io->module != NULL)
        source_io->module->Notify(msg_verbose, "    Deleting Connection.\n");
    delete next;
}

void
Connection::Propagate(long tick)
{
    if (delay == 0)
        return;
    // Return if both modules will not start in this tick - necessary when using threads
    if (tick % source_io->module->period != source_io->module->phase)
        return;
    if (tick % target_io->module->period != target_io->module->phase)
        return;
    if (source_io != NULL && source_io->module != NULL)
        source_io->module->Notify(msg_verbose, "  Propagating %s.%s -> %s.%s (%p -> %p) size = %d\n", source_io->module->GetName(), source_io->name, target_io->module->GetName(), target_io->name, source_io->data, target_io->data, size);
    for (int i=0; i<size; i++)
        target_io->data[0][i+target_offset] = source_io->data[delay-1][i+source_offset];
}

void
ThreadGroup::AddModule(Module * m)
{
    // Add module if thread is empty
    if (modules == NULL)
    {
        kernel->Notify(msg_verbose, "Adding module %s to new thread group\n", m->GetName());
        modules = m;
        last_module = m;
        period = m->period;
        phase = m->phase;
        return;
    }
    // Add module if this module should run in same thread as the other (and last) modules
    if (kernel->Preceedes(last_module, m))
    {
        kernel->Notify(msg_verbose, "Adding module %s to thread groups after %s\n", m->GetName(), last_module->GetName());
        last_module->next_in_threadGroup = m;
        last_module = m;
        return;
    }
    // Add to a new group if this was the last one
    if (next == NULL)
    {
        next = new ThreadGroup(kernel);
        next->AddModule(m);
        return;
    }
    // Try to add the module to the next group
    next->AddModule(m);
}

ThreadGroup::ThreadGroup(Kernel * k)
{
    kernel = k;
    next = NULL;
    modules = NULL;
    last_module = NULL;
    period = 1;
    phase = 0;
    thread = new Thread();
}

ThreadGroup::~ThreadGroup()
{}

static void *
ThreadGroup_Tick(void * group)
{
    ((ThreadGroup *)group)->Tick();
    return NULL;
}

void
ThreadGroup::Start(long tick)
{
    // Test if group should be started
    if (tick % period == phase)
    {
        if (thread->Create(ThreadGroup_Tick, (void*) this))
            printf("Thread Creation Failed!\n");
    }
}

void
ThreadGroup::Stop(long tick)
{
    // Test if group should be joined
    if ((tick + 1) % period == phase)
    {
        thread->Join();
    }
}

void
ThreadGroup::Tick()
{
    for (Module * m = modules; m != NULL; m = m->next_in_threadGroup)
    {
        m->timer->Restart();
        m->Tick();
        m->time += m->timer->GetTime();
        m->ticks += 1;
    }
}

Kernel::Kernel(Options * opt)
{
    options             = opt;
    useThreads          = options->GetOption('t') || options->GetOption('T');
    max_ticks           = string_to_int(options->GetArgument('s'), -1);
    tick_length         = string_to_int(options->GetArgument('r'), 0);
    print_mode		= print_normal;
    if (options->GetOption('q'))
        print_mode = print_silent;
    if (options->GetOption('v'))
        print_mode = print_verbose;
    tick                = 0;
    xmlDoc              = NULL;
    classes             = NULL;
    modules             = NULL;
    connections         = NULL;
    module_count        = 0;
    period_count        = 0;
    phase_count         = 0;
    end_of_file_reached = false;
    fatal_error_occured	= false;
    terminate			= false;
    sizeChangeFlag      = false;
    threadGroups        = NULL;
    
    logfile     = NULL;
    timer		= new Timer();

    // Compute ikaros root path
  
    ikaros_dir = NULL;
    if(is_absolute_path(IKAROSPATH))
        ikaros_dir = IKAROSPATH;
    else if(options->GetBinaryDirectory() != NULL)
        ikaros_dir = create_formatted_string("%s%s", options->GetBinaryDirectory(), IKAROSPATH);
    else
        Notify(msg_fatal_error, "The Ikaros root directory could not be established. Please set an absolute IKAROSPATH in IKAROS_System.h\n");
    
    // Compute ikc path and name
    
    ikc_dir = options->GetFileDirectory();
    ikc_file_name =  options->GetFileName();
}

Kernel::~Kernel()
{
    Notify(msg_verbose, "Deleting Kernel.\n");
    Notify(msg_verbose, "  Deleting Connections.\n");
    delete connections;
    Notify(msg_verbose, "  Deleting Modules.\n");
    delete modules;
    Notify(msg_verbose, "  Deleting Classes.\n");
    delete classes;
    Notify(msg_verbose, "Deleting Kernel Complete.\n");
    if (logfile) fclose(logfile);
}

void
Kernel::AddClass(char * name, ModuleCreator mc, char * path)
{
    if (path == NULL)
    {
        Notify(msg_warning, "Path to ikc file is missing for class \"%s\".\n", name);
        classes = new Class(name, mc, NULL, classes); // Add class anyway
        return;
    }
    char * path_to_ikc_file = create_formatted_string("%s%s%s.ikc", ikaros_dir, path, name);
    classes = new Class(name, mc, path_to_ikc_file, classes);
}

bool
Kernel::Terminate()
{
    if (max_ticks != -1 && tick >= max_ticks)
    {
        Notify(msg_verbose, "Max ticks reached.\n");
        return true;
    }
    return end_of_file_reached || fatal_error_occured  || global_fatal_error || terminate || global_terminate;
}

void
Kernel::Run()
{
    if (fatal_error_occured || global_fatal_error)
    {
        Notify(msg_fatal_error, "Terminating because a fatal error occured.\n");
        throw 4;
    }
    if (max_ticks == 0)
        return;
    Notify(msg_print, "Start\n");
    timer->Restart();
    while (!Terminate())
    {
        Tick();
        if (tick_length > 0)
        {
            float lag = timer->WaitUntil(float(tick*tick_length));
            if (lag > 0.1) Notify(msg_warning, "Lagging %.2f ms at tick = %ld\n", lag, tick);
        }
    }
}

void
Kernel::PrintTiming()
{
    total_time = timer->GetTime()/1000; // in seconds
    if (max_ticks != 0)
        Notify(msg_print, "Stop (%ld ticks, %.2f s, %.2f ticks/s, %.3f s/tick)\n", tick, total_time, float(tick)/total_time, total_time/float(tick));
}

void
Kernel::Propagate()
{
    for (Connection * c = connections; c != NULL; c = c->next)
        c->Propagate(tick);
}

void
Kernel::CheckInputs()
{
    for (Module * m = modules; m != NULL; m = m->next)
    {
        for (Module_IO * i = m->input_list; i != NULL; i = i->next)
            if (i->size == unknown_size)
            {
                // Check if connected
                bool connected = false;
                for (Connection * c = connections; c != NULL; c = c->next)
                    if (c->target_io == i)
                        connected = true;
                if (connected)
                {
                    Notify(msg_fatal_error, "Size of input \"%s\" of module \"%s\" (%s) could not be resolved.\n", i->name, i->module->name, i->module->GetClassName());
                }
                else
                    i->size = 0; // ok if not connected
            }
    }
}

void
Kernel::CheckOutputs()
{
    for (Module * m = modules; m != NULL; m = m->next)
    {
        for (Module_IO * i = m->output_list; i != NULL; i = i->next)
            if (i->size == unknown_size)
            {
                // Check if connected
                bool connected = false;
                for (Connection * c = connections; c != NULL; c = c->next)
                    if (c->source_io == i)
                        connected = true;
                if (connected)
                {
                    Notify(msg_fatal_error, "Size of output \"%s\" of module \"%s\" (%s) could not be resolved.\n", i->name, i->module->name, i->module->GetClassName());
                }
            }
    }
}

void
Kernel::InitInputs()
{
    for (Connection * c = connections; c != NULL; c = c->next)
    {
        if (c->source_io->size == unknown_size)
        {
            Notify(msg_fatal_error, "Output \"%s\" of module \"%s\" (%s) has unknown size.\n", c->source_io->name, c->source_io->module->name, c->source_io->module->GetClassName());
        }
        else if (c->delay == 0)
        {
            Notify(msg_verbose, "Short-circuiting zero-delay connection from \"%s\" of module \"%s\" (%s)\n", c->source_io->name, c->source_io->module->name, c->source_io->module->GetClassName());
            // already connected to 0 or longer delay?
            if (c->target_io->data != NULL)
            {
                Notify(msg_fatal_error, "Failed to connect zero-delay connection from \"%s\" of module \"%s\" (%s) because target is already connected.\n", c->source_io->name, c->source_io->module->name, c->source_io->module->GetClassName());
            }
            else
            {
                c->target_io->data = new float * [1];
                c->target_io->data[0] = c->source_io->data[0];
                c->target_io->matrix = new float ** [1];
                c->target_io->matrix[0] = c->source_io->matrix[0];
                c->target_io->sizex = c->source_io->sizex;
                c->target_io->sizey = c->source_io->sizey;
                c->target_io->size = c->source_io->size;
                c->target_io->max_delay = 0;
            }
        }
        else
        {
            // Check that this connection does not interfere with zero-delay connection
            if (c->target_io->max_delay == 0)
            {
                Notify(msg_fatal_error, "Failed to connect  from \"%s\" of module \"%s\" (%s) because target is already connected with zero-delay.\n", c->source_io->name, c->source_io->module->name, c->source_io->module->GetClassName());
            }
            // First connection to this target: initialize
            if (c->target_io->size == unknown_size)	// start calculation with size 0
                c->target_io->size = 0;
            int target_offset = c->target_io->size;
            c->target_io->size += c->source_io->size;
            // Target not used previously: ok to connect anything
            if (c->target_io->sizex == unknown_size)
            {
                Notify(msg_verbose, "New connection.\n");
                c->target_io->sizex = c->source_io->sizex;
                c->target_io->sizey = c->source_io->sizey;
            }
            // Connect one dimensional output
            else if (c->target_io->sizey == 1 && c->source_io->sizey == 1)
            {
                Notify(msg_verbose, "Adding additional connection.\n");
                c->target_io->sizex = c->target_io->size;
            }
            // Collapse matrix to array
            else
            {
                Notify(msg_warning, "Multiple connections to \"%s.%s\" with different no of rows. Input flattened.\n", c->target_io->module->name, c->target_io->name);
                c->target_io->sizex = c->target_io->size;
                c->target_io->sizey = 1;
            }
            // Set connection variables
            c->target_offset = target_offset;
            c->size = c->source_io->size;
            // Allocate input memory and reset
            Notify(msg_verbose, "Allocating memory for input \"%s\" of module \"%s\" with size %d (%dx%d).\n", c->target_io->name, c->target_io->module->name, c->target_io->size, c->target_io->sizex, c->target_io->sizey);
            c->target_io->SetSize(c->target_io->sizex, c->target_io->sizey);
            c->target_io->Allocate();
        }
    }
}

void
Kernel::InitOutputs()
{
    do
    {
        sizeChangeFlag = false;
        for (Module * m = modules; m != NULL; m = m->next)
            m->SetSizes();
        if (sizeChangeFlag)
            Notify(msg_verbose, "InitOutput: Iteration with changes\n");
        else
            Notify(msg_verbose, "InitOutput: Iteration with no changes\n");
    }
    while (sizeChangeFlag);
}

void
Kernel::AllocateOutputs()
{
    for (Module * m = modules; m != NULL; m = m->next)
        m->AllocateOutputs();
}

void
Kernel::InitModules()
{
    for (Module * m = modules; m != NULL; m = m->next)
        m->Init();
}

void
Kernel::NotifySizeChange()
{
    sizeChangeFlag = true;
}

void
Kernel::Init()
{
    if (options->GetFilePath())
        ReadXML(options->GetFilePath());
    else
        Notify(msg_warning, "No XML file supplied (Press CTRL-C to stop).\n");
    SortModules();
    CalculateDelays();
    InitOutputs();      // Calculate the output sizes for outputs that have not been specified at creation
    AllocateOutputs();
    InitInputs();		// Calculate the input sizes and allocate memory for the inputs or connect 0-delays
    CheckOutputs();
    CheckInputs();
    if (fatal_error_occured)
        return;
    InitModules();
}

void
Kernel::Tick()
{
    Notify(msg_verbose, "Kernel::Tick()\n");
    Propagate();
    DelayOutputs();
    if (useThreads)
    {
        for (ThreadGroup * g = threadGroups; g != NULL; g = g->next)
            g->Start(tick);
        for (ThreadGroup * g = threadGroups; g != NULL; g = g->next)
            g->Stop(tick);
    }
    else if (print_mode < msg_verbose)
    {
        for (Module * m = modules; m != NULL; m = m->next)
            if (tick % m->period == m->phase)
            {
                m->timer->Restart();
                m->Tick();
                m->time += m->timer->GetTime();
                m->ticks += 1;
            }
    }
    else
    {
        for (Module * m = modules; m != NULL; m = m->next)
            if (tick % m->period == m->phase)
            {
                m->timer->Restart();
                Notify(msg_verbose, "%s::Tick (%s) Start\n", m->GetName(), m->GetClassName());
                m->Tick();
                Notify(msg_verbose, "%s::Tick (%s) End\n", m->GetName(), m->GetClassName());
                m->time += m->timer->GetTime();
                m->ticks += 1;
            }

    }
    tick++;
}

void
Kernel::DelayOutputs()
{
    for (Module * m = modules; m != NULL; m = m->next)
        m->DelayOutputs();
}

void
Kernel::AddModule(Module * m)
{
    if (!m) return;
    m->next = modules;
    modules = m;
    m->kernel = this;
}

Module *
Kernel::GetModule(const char * n)
{
    for (Module * m = modules; m != NULL; m = m->next)
        if (equal_strings(n, m->name))
            return m;
    return NULL;
}

bool
Kernel::GetSource(XMLElement * group, Module * &m, Module_IO * &io, char * source_module_name, char * source_name)
{
    for (XMLNode * xml = group->content; xml != NULL; xml = xml->next)
    {
        if (xml->IsElement())
        {
            if (equal_strings(((XMLElement *)(xml))->name, "module"))
            {
                char * name = ((XMLElement *)(xml))->GetAttribute("name");
                if (equal_strings(name, source_module_name))
                {
                    m = (Module *)(xml->aux);
                    if (m != NULL)
                        io = m->GetModule_IO(m->output_list, source_name);
                    if (m != NULL && io != NULL)
                        return true;
                    return false;
                }
            }
            else if (equal_strings(((XMLElement *)(xml))->name, "group"))
            {
                char * name = ((XMLElement *)(xml))->GetAttribute("name");
                if (equal_strings(name, source_module_name))
                {
                    // Translate output name
                    for (XMLElement * output = ((XMLElement *)(xml))->GetElement("output"); output != NULL; output = output->next->FindElement("output"))
                    {
                        char * n = output->GetAttribute("name");
                        if (n == NULL)
                            return false;
                        if (equal_strings(n, source_name)  || equal_strings(n, "*"))
                        {
                            char * new_module = output->GetAttribute("sourcemodule");
                            char * new_source = output->GetAttribute("source");
                            if (new_source == NULL)
                                new_source = source_name;	// retain name
                            return GetSource(((XMLElement *)(xml)), m, io, new_module, new_source);
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool
Kernel::GetTarget(XMLElement * group, Module * &m, Module_IO * &io, char * target_module_name, char * target_name)
{
    for (XMLNode * xml = group->content; xml != NULL; xml = xml->next)
    {
        if (xml->IsElement())
        {
            if (equal_strings(((XMLElement *)(xml))->name, "module"))
            {
                char * name = ((XMLElement *)(xml))->GetAttribute("name");
                if (equal_strings(name, target_module_name))
                {
                    //	printf("Module found %s\n", name);
                    m = (Module *)(xml->aux);
                    if (m != NULL)
                        io = m->GetModule_IO(m->input_list, target_name);
                    if (m != NULL && io != NULL)
                        return true;
                    return false;
                }
            }
            else if (equal_strings(((XMLElement *)(xml))->name, "group"))
            {
                char * name = ((XMLElement *)(xml))->GetAttribute("name");
                if (equal_strings(name, target_module_name))
                {
                    // Translate input name
                    for (XMLElement * output = ((XMLElement *)(xml))->GetElement("input"); output != NULL; output = output->next->FindElement("input"))
                    {
                        char * n = output->GetAttribute("name");
                        if (n == NULL)
                            return false;
                        if (equal_strings(n, target_name) || equal_strings(n, "*"))
                        {
                            char * new_module = output->GetAttribute("targetmodule");
                            char * new_target = output->GetAttribute("target");
                            if (new_target == NULL)
                                new_target = target_name;	// retain name
                            return GetTarget(((XMLElement *)(xml)), m, io, new_module, new_target);
                        }
                    }
                }
            }
        }
    }
    return false;
}

void
Kernel::Connect(const char * source_module_name, const char * source_output_name, const char * taget_module_name, const char * target_input_name, int d)
{
    // Check that connection is possible
    Module * source_module = GetModule(source_module_name);
    if (source_module == NULL)
    {
        Notify(msg_warning, "Could not connect: Module \"%s\" does not exist.\n", source_module_name);
        return;
    }
    Module_IO * source_output = source_module->GetModule_IO(source_module->output_list, source_output_name);
    if (source_output == NULL)
    {
        Notify(msg_warning, "Could not connect: Output \"%s\" of module \"%s\" does not exist.\n", source_output_name, source_module_name);
        return;
    }
    Module * target_module = GetModule(taget_module_name);
    if (target_module == NULL)
    {
        Notify(msg_warning, "Could not connect: Module \"%s\" does not exists.\n", taget_module_name);
        return;
    }
    Module_IO * target_input = target_module->GetModule_IO(target_module->input_list, target_input_name);
    if (target_input == NULL)
    {
        Notify(msg_warning, "Could not connect: Input \"%s\" of module \"%s\" does not exist.\n", target_input_name, taget_module_name);
        return;
    }
    // Make the connection

    connections = new Connection(connections, source_output, 0, target_input, unknown_size, unknown_size, d);  // Do not set sizes yet
}

bool
Kernel::Preceedes(Module * a, Module * b)
{
    // Base case
    for (Connection * c = connections; c != NULL; c = c->next)
        if (c->delay == 0 && c->source_io->module == a && c->target_io->module == b)
            return true;
    // Transitivity test
    for (Connection * c = connections; c != NULL; c = c->next)
        if (c->delay == 0 && c->source_io->module == a && Preceedes(c->target_io->module,  b))
            return true;
    return false;
}

int
Kernel::CalculateInputSize(Module_IO * i)
{
    // The input size has not yet been determined; use the connection list to calculate it
    // Scan through the connection list to find the size of the input
    // The size is the sum of the sizes of all outputs connected to this input
    // The input size is unspecified if one of the outputs has unknown size
    int s = 0;
    for (Connection * c = connections; c != NULL; c = c->next)
    {
        if (c->target_io == i)
        {
            if (c->source_io->size == unknown_size)
                return unknown_size;
            else
                s += c->source_io->size;
        }
    }
    return s;
}

int
Kernel::CalculateInputSizeX(Module_IO * i)
{
    // The input size has not yet been determined; use the connection list to calculate it
    // Scan through the connection list to find the size of the input
    // For matrices, there can only be one input
    // The input size is unspecified if one of the outputs has unknown size
    int s = 0;
    for (Connection * c = connections; c != NULL; c = c->next)
    {
        if (c->target_io == i)
        {
            if (c->source_io->sizex == unknown_size)
                return unknown_size;
            else if (s == 0)
                s =  c->source_io->sizex;
            else
                return CalculateInputSize(i);
        }
    }
    return s;
}

int
Kernel::CalculateInputSizeY(Module_IO * i)
{
    // The input size has not yet been determined; use the connection list to calculate it
    // Scan through the connection list to find the size of the input
    // For matrices, there can only be one input
    // The input size is unspecified if one of the outputs has unknown size
    int s = 0;
    for (Connection * c = connections; c != NULL; c = c->next)
    {
        if (c->target_io == i)
        {
            if (c->source_io->sizey == unknown_size)
                return unknown_size;
            else if (s == 0)
                s =  c->source_io->sizey;
            else
                return 1;
        }
    }
    return s;
}

void
Kernel::SortModules()
{
    // For statistics only
    for (Module * m = modules; m != NULL; m = m->next)
    {
        phase_count = (m->phase < phase_count ? phase_count : m->phase);
        period_count = (m->period < period_count ? period_count : m->period);
        module_count++;
    }
    if (phase_count > period_count)
        period_count = phase_count;
    // Build a new sorted list of modules (precedence order) using selection sort
    Module * sorted_modules = NULL;
    while (modules != NULL)
    {
        // Find smallest module
        Module * sm = modules;
        for (Module * m = modules; m != NULL; m = m->next)
        {
            if (Preceedes(sm, m))
                sm = m;
        }
        // Remove from list
        if (sm == modules)  // First
        {
            modules = modules->next;
            sm->next = sorted_modules;
            sorted_modules = sm;
        }
        else
        {
            // Find prev (double linked list would have been better)
            Module * psm = NULL;
            for (psm = modules; psm->next != sm; psm = psm->next)
                ;
            psm->next = sm->next;
            sm->next = sorted_modules;
            sorted_modules = sm;
        }
    }
    modules = sorted_modules;
    // Check for loops
    for (Module * m = modules; m != NULL; m = m->next)
        if (Preceedes(m, m))
            Notify(msg_fatal_error, "Module \"%s\" (%s) has a zero-delay connection to itself.\n", m->GetName(), m->GetClassName());
    // Create Thread Groups
    threadGroups = new ThreadGroup(this);
    for (Module * m = modules; m != NULL; m = m->next)
        threadGroups->AddModule(m);
}

void
Kernel::CalculateDelays()
{
    for (Connection * c = connections; c != NULL; c = c->next)
    {
        if (c->delay > c->source_io->max_delay)
            c->source_io->max_delay = c->delay;
    }
}

bool
Kernel::InputConnected(Module * m, char * input_name)
{
    if (connections == NULL)
        return false;
    for (Connection * c = connections; c != NULL; c = c->next)
        if (c->target_io->module == m && equal_strings(input_name, c->target_io->name))
            return true;
    return false;
}

bool
Kernel::OutputConnected(Module * m, char * output_name)
{
    if (connections == NULL)
        return false;
    for (Connection * c = connections; c != NULL; c = c->next)
        if (c->source_io->module == m && equal_strings(output_name, c->source_io->name))
            return true;
    return false;
}

void
Kernel::ListInfo()
{
    if (!options->GetOption('i')) return;
    Notify(msg_print, "\n");
    Notify(msg_print, "Ikaros version %s\n", VERSION);
    Notify(msg_print, "\n");
    Notify(msg_print, "%s\n", PLATFORM);
#ifdef POSIX
    Notify(msg_print, "POSIX\n");
#endif
#ifdef USE_BSD_SOCKET
    Notify(msg_print, "BSD-socket\n");
#endif
#ifdef USE_WIN_SOCKET
    Notify(msg_print, "WIN-socket\n");
#endif
#ifdef USE_LIBJPEG
    Notify(msg_print, "libjpeg\n");
#endif
#ifdef USE_THREADS
    Notify(msg_print, "threads\n");
#endif
#ifdef USE_BLAS
    Notify(msg_print, "BLAS\n");
#endif
#ifdef USE_QUICKTIME
    Notify(msg_print, "Quicktime\n");
#endif
#ifdef USE_VIMAGE
    Notify(msg_print, "vImage\n");
#endif
#ifdef USE_VFORCE
    Notify(msg_print, "vForce\n");
#endif
#ifdef USE_VDSP
    Notify(msg_print, "vDSP\n");
#endif
#ifdef USE_VDSP
    Notify(msg_print, "GSL\n");
#endif
    Notify(msg_print, "\n");
    Notify(msg_print, "ikc file name: %s\n", ikc_file_name);
    Notify(msg_print, "ikc file directory: %s\n", ikc_dir);
    Notify(msg_print, "ikaros root directory: %s\n", ikaros_dir);
}

void
Kernel::ListModulesAndConnections()
{
    if (!options->GetOption('m') && !options->GetOption('a')) return;
    Notify(msg_print, "\n");
    Notify(msg_print, "Modules:\n");
    Notify(msg_print, "\n");
    for (Module * m = modules; m != NULL; m = m->next)
    {
        Notify(msg_print, "  %s (%s) [%d, %d]:\n", m->name, m->class_name, m->period, m->phase);
        for (Module_IO * i = m->input_list; i != NULL; i = i->next)
            Notify(msg_print, "    %-10s\t(Input) \t%6d%6d%12p\n", i->name, i->sizex, i->sizey, (i->data == NULL ? NULL : i->data[0]));
        for (Module_IO * i = m->output_list; i != NULL; i = i->next)
            Notify(msg_print, "    %-10s\t(Output)\t%6d%6d%12p\t(%d)\n", i->name, i->sizex, i->sizey, (i->data == NULL ? NULL : i->data[0]), i->max_delay);
        Notify(msg_print, "\n");
    }
    Notify(msg_print, "Connections:\n");
    Notify(msg_print, "\n");
    for (Connection * c = connections; c != NULL; c = c->next)
        if (c->delay == 0)
            Notify(msg_print, "  %s.%s[%d..%d] == %s.%s[%d..%d] (%d)\n",
                   c->source_io->module->name, c->source_io->name, 0, c->source_io->size-1,
                   c->target_io->module->name, c->target_io->name, 0, c->source_io->size-1,
                   c->delay);
        else if (c->size > 1)
            Notify(msg_print, "  %s.%s[%d..%d] -> %s.%s[%d..%d] (%d)\n",
                   c->source_io->module->name, c->source_io->name, c->source_offset, c->source_offset+c->size-1,
                   c->target_io->module->name, c->target_io->name, c->target_offset, c->target_offset+c->size-1,
                   c->delay);
        else
            Notify(msg_print, "  %s.%s[%d] -> %s.%s[%d] (%d)\n",
                   c->source_io->module->name, c->source_io->name, c->source_offset,
                   c->target_io->module->name, c->target_io->name, c->target_offset,
                   c->delay);
    Notify(msg_print, "\n");
}

void
Kernel::ListThreads()
{
    if (!options->GetOption('T') && !(options->GetOption('a') && options->GetOption('t'))) return;
    Notify(msg_print, "\n");
    Notify(msg_print,"ThreadManagers:\n");
    Notify(msg_print, "\n");
    int tt = 0;
    for (ThreadGroup * t = threadGroups; t != NULL; t = t->next)
    {
        Notify(msg_print,"ThreadManager %d [Period:%d, Phase:%d]\n", tt++, t->period, t->phase);
        for (Module * m = t->modules; m != NULL; m=m->next_in_threadGroup)
            Notify(msg_print,"\tModule: %s\n", m->GetName());
    }
    Notify(msg_print, "\n");
}

void
Kernel::ListScheduling()
{
    if (!options->GetOption('l') && !options->GetOption('a')) return;
    Notify(msg_print, "Scheduling:\n");
    Notify(msg_print, "\n");
    for (int t=0; t<period_count; t++)
    {
        int tm = 0;
        for (Module * m = modules; m != NULL; m = m->next)
            if (t % m->period == m->phase)
                Notify(msg_print,"  %02d.%02d: %s (%s)\n", t, tm++, m->GetName(), m->GetClassName());
    }

    Notify(msg_print, "\n");
    Notify(msg_print, "\n");
}

void
Kernel::ListClasses()
{
    if (!options->GetOption('c')) return;
    int i = 0;
    Notify(msg_print, "\n");
    Notify(msg_print, "Classes:\n");
    for (Class * c = classes; c != NULL; c = c->next)
    {
        Notify(msg_print, "\t%s\n", c->name);
        i++;
    }
    Notify(msg_print, "No of modules: %d.\n", i);
}

void
Kernel::ListProfiling()
{
    if (!options->GetOption('p')) return;
    // Calculate Total Time
    float total_module_time = 0;
    for (Module * m = modules; m != NULL; m = m->next)
        total_module_time += m->time;
    Notify(msg_print, "\n");
    Notify(msg_print, "Time (ms):\n");
    Notify(msg_print, "-------------------\n");
    Notify(msg_print, "Modules: %10.2f\n", total_module_time);
    Notify(msg_print, "Other:   %10.2f\n", 1000*total_time-total_module_time);
    Notify(msg_print, "-------------------\n");
    Notify(msg_print, "Total:   %10.2f\n", 1000*total_time);
    Notify(msg_print, "\n");
    if (total_module_time == 0)
        return;
    Notify(msg_print, "Time in Each Module:\n");
    Notify(msg_print, "\n");
    Notify(msg_print, "%-20s%-20s%10s%10s%10s\n", "Module", "Class", "Count", "Avg (ms)", "Time %");
    Notify(msg_print, "----------------------------------------------------------------------\n");
    for (Module * m = modules; m != NULL; m = m->next)
        if (m->ticks > 0)
            Notify(msg_print, "%-20s%-20s%10.0f%10.2f%10.1f\n", m->GetName(), m->GetClassName(), m->ticks, (m->time/m->ticks), 100*(m->time/total_module_time));
        else
            Notify(msg_print, "%-20s%-20s%       ---f\n", m->GetName(), m->GetClassName());
    Notify(msg_print, "----------------------------------------------------------------------\n");
    if (useThreads)
        Notify(msg_print, "Note: Time is real-time, not time in thread.\n");
    Notify(msg_print, "\n");
}

void
Kernel::Notify(int msg, char *format, ...)
{
    switch (msg)
    {
    case msg_fatal_error:
        fatal_error_occured		= true;
        break;
    case msg_end_of_file:
        end_of_file_reached	= true;
        break;
    case msg_terminate:
        terminate				= true;
        break;
    default:
        break;
    }
    if (msg > print_mode)
        return;
    char 	message[512];
    int n = 0;
    switch (msg)
    {
    case msg_fatal_error:
        n = snprintf(message, 512, "ERROR: ");
        break;
    case msg_end_of_file:
        n = snprintf(message, 512, "END-OF-FILE: ");
        break;
    case msg_terminate:
        n = snprintf(message, 512, "TERMINATE. ");
        break;
    case msg_warning:
        n = snprintf(message, 512, "WARNING: ");
        break;
    default:
        break;
    }
    va_list 	args;
    va_start(args, format);
    vsnprintf(&message[n], 512, format, args);
    va_end(args);
    printf("IKAROS: %s", message);
    if (logfile != NULL)
        fprintf(logfile, "%5ld: %s", tick, message);	// Print in both places
}

//
//
//	New XML Parser including IKC support
//
//
void
Kernel::BuildGroup(XMLElement * group_xml, char * current_class)
{
    if (group_xml->GetElement("views"))
        Notify(msg_warning, "The <views> tag should not be used in IKC files.");
    // Start processing the XML tree
    for (XMLNode * xml_node = group_xml->content; xml_node != NULL; xml_node = xml_node->next)
        if (xml_node->IsElement() && equal_strings(((XMLElement *)(xml_node))->name, "module"))	// add module
        {
            char * class_name = ((XMLElement *)(xml_node))->GetAttribute("class");
            char * module_name = ((XMLElement *)(xml_node))->GetAttribute("name");	// only for backward compatibility
            bool exists = false;
            char filename[PATH_MAX] ="";
            if (!equal_strings(class_name, current_class))  // Check that we are not in a class file
            {
                // Check if class is defined externally in IKC directory
                strcat(filename, class_name);
                strcat(filename, ".ikc");
                FILE * t = fopen(filename, "r");
                exists = (t != NULL);
                if (t) fclose(t);
                // Otherwise search among classes
                if (!exists)
                {
                    char * class_path = classes->GetClassPath(class_name);
                    if (class_path != NULL)
                    {
                        filename[0] = 0;
                        strcpy(filename, class_path);
                        FILE * t = fopen(filename, "r");
                        exists = (t != NULL);
                        if (t) fclose(t);
                        if (!exists)
                            Notify(msg_warning, "Class file \"%s\" does not exist.\n", filename);
                    }
                }
            }
            if (exists) 
            {
                XMLDocument * cDoc = new XMLDocument(filename);
                XMLElement * cgroup = cDoc->xml->FindElement("group");
                // Merge the two trees

                // 0. Find previous node
                XMLElement * p = (XMLElement *)((XMLElement *)(xml_node)->parent);
                XMLNode * prev = NULL;
                if (p->content != xml_node)
                {
                    prev = p->content;
                    while (prev->next != NULL && prev->next != xml_node)
                        prev=prev->next;
                }

                // 1. Replace the module element with the group element of the included file
                cgroup->next = xml_node->next;
                if (prev == NULL)
                    p->content = cgroup;
                else
                    prev->next = cgroup;

                // 2. Copy attributes
                cgroup->attributes = ((XMLElement *)(xml_node))->attributes;
                ((XMLElement *)(xml_node))->attributes = NULL;

                // 3. Copy elements

                XMLNode * last = cgroup->content;
                if (last != NULL)
                {
                    while (last->next != NULL)
                        last = last->next;
                    last->next = ((XMLElement *)(xml_node))->content;
                }
                else
                {
                    cgroup->content = ((XMLElement *)(xml_node))->content;
                }

                ((XMLElement *)(xml_node))->content = NULL;

                // Build the class group
                BuildGroup(cgroup, class_name);
            }
            else // Create the module using standard class
            {
                if (class_name  != NULL)	// Create the module
                {
                    Parameter * parameter = new Parameter(this, ((XMLElement *)(xml_node)));
                    Module * m = CreateModule(classes, class_name, module_name, parameter);
                    delete parameter;
                    if (m == NULL)
                        Notify(msg_warning, "Could not create module: Class \"%s\" does not exist.\n", class_name);
                    if (useThreads && m->phase != 0)
                        Notify(msg_fatal_error, "phase != 0 not yet supported in threads.");
                    xml_node->aux = (void *)m;
                    AddModule(m);
                }
            }
        }
        else if (xml_node->IsElement() && equal_strings(((XMLElement *)(xml_node))->name, "group"))	// add group
        {
            BuildGroup((XMLElement *)(xml_node));
        }
    // Create connections in group
    for (XMLElement * xml_connection = group_xml->GetElement("connection"); xml_connection != NULL; xml_connection = xml_connection->next->FindElement("connection"))
    {
        char * sm_name = xml_connection->GetAttribute("sourcemodule");
        char * s_name = xml_connection->GetAttribute("source");
        char * tm_name = xml_connection->GetAttribute("targetmodule");
        char * t_name = xml_connection->GetAttribute("target");
        int d = string_to_int(xml_connection->GetAttribute("delay"), 1);

        Module * sm;
        Module_IO * sio;
        if (!GetSource(group_xml, sm, sio, sm_name, s_name))
        {
            Notify(msg_fatal_error, "Connection source %s.%s not found.\n", sm_name, s_name);
            return;
        }
        Module * tm;
        Module_IO * tio;
        if (!GetTarget(group_xml, tm, tio, tm_name, t_name))
        {
            Notify(msg_fatal_error, "Connection target %s.%s not found.\n", tm_name, t_name);
            return;
        }
        connections = new Connection(connections, sio, 0, tio, unknown_size, unknown_size, d);  // Do not set sizes yet
    }
}

void
Kernel::ReadXML(char * filename)
{
    if (ikc_file_name[0] == 0)
    {
        Notify(msg_fatal_error, "Empty file name.\n");
        return;
    }
    char path[PATH_MAX];
    strcpy(path, ikc_dir);
    strcat(path, ikc_file_name);
    Notify(msg_print, "Reading XML file \"%s\".\n", ikc_file_name);
    if (chdir(ikc_dir) < 0)
    {
        Notify(msg_fatal_error, "The directory \"%s\" could not be found.\n", ikc_dir);
        return;
    }
    xmlDoc = new XMLDocument(ikc_file_name, options->GetOption('X'));
    if (xmlDoc->xml == NULL)
    {
        Notify(msg_fatal_error, "Could not read (or find) \"%s\".\n", ikc_file_name);
        return;
    }
    XMLElement * xml = xmlDoc->xml->FindElement("group");
    if (xml == NULL)
    {
        Notify(msg_warning, "Did not find <group> element in IKC/XML file \"%s\". Will use legacy mode.\n", ikc_file_name);
        ReadCompatibilityXML();
        return;
    }
    // Build The Main Group
    BuildGroup(xml);
    if (options->GetOption('x'))
        xmlDoc->Print(stdout);
}

//
//	Use compatibility mode
//
void
Kernel::ReadCompatibilityXML()
{
    XMLElement * xml = xmlDoc->xml->FindElement("network");
    if (xml == NULL)
    {
        Notify(msg_fatal_error, "Could not find <network> tag in XML file.\n");
        return;
    }
    // Create Modules
    XMLElement * xml_modules = xml->GetElement("modules");
    if (xml_modules == NULL)
    {
        Notify(msg_warning, "No modules.");
        return;
    }
    for (XMLElement * xml_element = xml_modules->GetElement("module"); xml_element != NULL; xml_element = xml_element->next->FindElement("module"))
    {
        char * class_name = create_string(xml_element->GetAttribute("class"));
        char * module_name = create_string(xml_element->GetAttribute("name"));
        if (class_name  != NULL && module_name != NULL)
        {
            Module * m = CreateModule(classes, class_name, module_name, new Parameter(this, xml_element));
            if (m == NULL)
                Notify(msg_warning, "Could not create module: Class \"%s\" does not exist.\n", class_name);
            if (useThreads && m->phase != 0)
                Notify(msg_fatal_error, "phase != 0 not yet supported in threads.");
            xml_element->aux = (void *)m;
            AddModule(m);
        }
    }
    // Create Connections
    XMLElement * xml_connections = xml->GetElement("connections");
    if (xml_connections == NULL)
    {
        Notify(msg_warning, "No connections.\n");
        return;
    }
    XMLElement * xml_first_connection = xml_connections->GetElement("connection");
    if (xml_first_connection == NULL)
    {
        Notify(msg_warning, "No connections.\n");
        return;
    }
    for (XMLElement * xml_connection = xml_connections->GetElement("connection"); xml_connection != NULL; xml_connection = xml_connection->next->FindElement("connection"))
        Connect(xml_connection->GetAttribute("sourcemodule"), xml_connection->GetAttribute("source"), xml_connection->GetAttribute("targetmodule"), xml_connection->GetAttribute("target"), string_to_int(xml_connection->GetAttribute("delay"), 1));
    if (options->GetOption('x'))
        xmlDoc->Print(stdout);
}

