//
//	IKAROS_Math.cc		Various math functions for IKAROS
//
//    Copyright (C) 2006-2007  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    See http://www.ikaros-project.org/ for more information.
//

#include "IKAROS_System.h"
#include "IKAROS_Math.h"
#include "IKAROS_Utils.h"
#include <cmath>
#include <cstdlib>

#ifdef MAC_OS_X
#include <Accelerate/Accelerate.h>
#endif

#ifdef LINUX
#define	MAXFLOAT	3.402823466e+38f
#ifdef USE_BLAS
extern "C"
{
#include "gsl/gsl_blas.h"
}
#endif
#endif

#ifdef WINDOWS
#define	MAXFLOAT	3.402823466e+38f
#ifdef USE_BLAS
extern "C"
{
#include "gsl/gsl_blas.h"
}
#endif
#endif

#ifdef WINDOWS32
#define M_PI 3.14159265358979323846264338328f
#endif

namespace ikaros
{
extern const float maxfloat = MAXFLOAT;
extern const float pi = M_PI;
extern const float sqrt2pi = sqrt(2*pi);
// misc scalar functions
float trunc(float x)
{
#ifdef WINDOWS32
    return (float)((int)x);
#else
    return ::truncf((int)x);
#endif
}
float exp(float x)
{
    return ::expf(x);
}
float pow(float x, float y)
{
    return ::powf(x, y);
}
float log(float x)
{
    return ::logf(x);
}
float log10(float x)
{
    return ::log10f(x);
}
float	sin(float x)
{
    return ::sinf(x);
}
float	cos(float x)
{
    return ::cosf(x);
}
float	tan(float x)
{
    return ::tanf(x);
}
float	asin(float x)
{
    return ::asinf(x);
}
float	acos(float x)
{
    return ::acosf(x);
}
float	atan(float x)
{
    return ::atanf(x);
}
float	atan2(float x, float y)
{
    return ::atan2f(x, y);
}
float *
atan2(float * r, float * a, float * b, int size)
{
#ifdef USE_VFORCE
    vvatan2f(r, a, b, &size);
#else
    for (int i=0; i<size; i++)
        r[i] = ::atan2f(a[i], b[i]);
#endif
    return r;
}

float **
atan2(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    atan2(*r, *a, *b, sizex*sizey);
    return r;
}

// gaussian [not optimized]
float
gaussian(float x, float sigma)
{
    return (1/(sigma*sqrt2pi))*::expf(-sqr(x)/(2*sqr(sigma)));
}

float *
gaussian(float * r, float * a, float sigma, int size)
{
    for (int i=0; i<size; i++)
        r[i] = gaussian(a[i], sigma);
    return r;
}

float **
gaussian(float ** r, float ** m, float sigma, int sizex, int sizey)
{
    gaussian(*r, *m, sigma, sizex*sizey);
    return r;
}

float gaussian1(float x, float sigma)
{
    return ::expf(-sqr(x)/(2*sqr(sigma)));
}

float *
gaussian1(float * r, float * a, float sigma, int size)
{
    for (int i=0; i<size; i++)
        r[i] = gaussian1(a[i], sigma);
    return r;
}

float **
gaussian1(float ** r, float ** m, float sigma, int sizex, int sizey)
{
    gaussian1(*r, *m, sigma, sizex*sizey);
    return r;
}

float *
gaussian(float * r, float x, float sigma, int size)
{
    float center = x;
    for (int i=0; i<size; i++)
        r[i] = gaussian(center-float(i), sigma);
    return r;
}

float **
gaussian(float ** r, float x, float y, float sigma, int sizex, int sizey)
{
    float center_x = x;
    float center_y = y;
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            r[j][i] = gaussian(hypot(center_x-float(i), center_y-float(j)), sigma);
    return r;
}

float *
gaussian1(float * r, float x, float sigma, int size)
{
    float center = x;
    for (int i=0; i<size; i++)
        r[i] = gaussian1(center-float(i), sigma);
    return r;
}

float **
gaussian1(float ** r, float x, float y, float sigma, int sizex, int sizey)
{
    float center_x = x;
    float center_y = y;
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            r[j][i] = gaussian1(hypot(center_x-float(i), center_y-float(j)), sigma);
    return r;
}


// random
float
random(float low, float high)
{
#ifdef WINDOWS
    return low + (float(::rand())/float(RAND_MAX))*(high-low);
#else
    return low + (float(::random())/float(RAND_MAX))*(high-low);
#endif
}


float *
random(float * r, float low, float high, int size)
{
    for (int i=0; i<size; i++)
        r[i] = random(low, high);
    return r;
}


float **
random(float ** r, float low, float high, int sizex, int sizey)
{
    random(*r, low, high, sizex*sizey);
    return r;
}

// sqr
int
sqr(int x)
{
    return x*x;
}

float
sqr(float x)
{
    return x*x;
}

float *
sqr(float * a, int size)
{
    sqr(a, a, size);
    return a;
}

float **
sqr(float ** a, int sizex, int sizey)
{
    sqr(*a, sizex*sizey);
    return a;
}

float *
sqr(float * r, float * a, int size)
{
#ifdef USE_VDSP
    vDSP_vsq(a, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = a[i] * a[i];
#endif
    return r;
}

float **
sqr(float ** r, float ** a, int sizex, int sizey)
{
    sqr(*r, *a, sizex*sizey);
    return r;
}

// tests
bool
zero(float * a, int size)
{
    for (int i=0; i<size; i++)
        if (a[i] != 0)
            return false;
    return true;
}

bool
zero(float ** a, int sizex, int sizey)
{
    return zero(*a, sizex*sizey);
}

bool
non_zero(float * a, int size)
{
    for (int i=0; i<size; i++)
        if (a[i] != 0)
            return true;
    return false;
}

bool
non_zero(float ** a, int sizex, int sizey)
{
    return non_zero(*a, sizex*sizey);
}

// sqrt
float
sqrt(int x)
{
    return ::sqrtf(float(x));
}

float
sqrt(float x)
{
    return ::sqrtf(x);
}

float *
sqrt(float * a, int size)
{
    for (int i=0; i<size; i++)
        a[i] = sqrt(a[i]);
    return a;
}

float **
sqrt(float ** a, int sizex, int sizey)
{
    sqrt(*a, sizex*sizey);
    return a;
}

float *
sqrt(float * r, float * a, int size)
{
    for (int i=0; i<size; i++)
        r[i] = sqrt(a[i]);
    return r;
}

float **
sqrt(float ** r, float ** a, int sizex, int sizey)
{
    sqrt(*r, *a, sizex*sizey);
    return r;
}

// hypot
float
hypot(float x, float y)
{
#ifdef WINDOWS32
    return (float)_hypot(x, y);
#else
    return ::hypotf(x, y);
#endif
}

float *
hypot(float * r, float * a, float * b, int size)
{
#ifdef USE_VDSP
    vDSP_vdist(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = hypot(a[i], b[i]);
#endif
    return r;
}

float **
hypot(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    hypot(*r, *a, *b, sizex*sizey);
    return r;
}

// distances & norms
float
dist(float * a, float * b, int size)
{
    float s = 0;
    for (int i=0; i < size; i++)
        s += (a[i] - b[i]) * (a[i] - b[i]);
    return sqrt(s);
}

float
dist(float ** a, float ** b, int sizex, int sizey)
{
    return dist(*a, *b, sizex*sizey);
}

float
dist1(float * a, float * b, int size)
{
    float r = 0;
    for (int i=0; i < size; i++)
        r += fabs(a[i] - b[i]);
    return r;
}

float
dist1(float ** a, float ** b, int sizex, int sizey)
{
    return dist1(*a, *b, sizex*sizey);
}

float
norm(float * a, int size)
{
#ifdef USE_BLAS
    return cblas_snrm2(size, a, 1);
#else
    float r = 0;
    for (int i=0; i < size; i++)
        r += a[i] * a[i];
    return sqrt(r);
#endif
}

float
norm(float ** a, int sizex, int sizey)
{
    return norm(*a, sizex*sizey);
}

float
norm1(float * a, int size)
{
#ifdef USE_BLAS
    return cblas_sasum(size, a, 1);
#else
    float r = 0;
    for (int i=0; i < size; i++)
        r += fabs(a[i]);
    return r;
#endif
}

float
norm1(float ** a, int sizex, int sizey)
{
    return norm1(*a, sizex*sizey);
}

float *
normalize(float * a, int size)
{
    float n = norm(a, size);
    if(n!=0)
       multiply(a, 1/n, size);
    return a;
}

float **
normalize(float ** m, int sizex, int sizey)
{
    normalize(*m, sizex*sizey);
    return m;
}

float *
normalize1(float * a, int size)
{
    float n = norm1(a, size);
    if(n!=0)
        multiply(a, 1/n, size);
    return a;
}

float **
normalize1(float ** m, int sizex, int sizey)
{
    normalize1(*m, sizex*sizey);
    return m;
}

float *
normalize_max(float * a, int size)
{
    float n = max(a, size);
    if(n!=0)
        multiply(a, 1/n, size);
    return a;
}

float **
normalize_max(float ** m, int sizex, int sizey)
{
    normalize_max(*m, sizex*sizey);
    return m;
}

// abs
int
abs(int x)
{
    return (x < 0 ? -x : x);
}

float
abs(float x)
{
    return (x < 0 ? -x : x);
}

float *
abs(float * a, int size)
{
//		vvfabf(r, a, &size); ***
    float t;
    for (int i=0; i<size; i++, a++)
        if ((t = *a) < 0)
            *a = -t;
    return a;
}

float **
abs(float ** a, int sizex, int sizey)
{
    abs(*a, *a, sizex*sizey);
    return a;
}

float *
abs(float * r, float * a, int size)
{
//		vvfabf(r, a, &size); ***
#ifdef USE_VDSP_NOT_YET
    vDSP_vabs(a, 1, r, 1, size);	// slower than scalar code on Intel Xeon
#else
    float t;
    for (int i=0; i<size; i++)
        if ((t = (*a++)) < 0)
            (*r++) = -t;
        else
            (*r++) = t;
#endif
    return r;
}

float **
abs(float ** r, float ** m, int sizex, int sizey)
{
    abs(*r, m[0], sizex*sizey);
    return r;
}

// min
int
min(int x, int y)
{
    return (x < y ? x : y);
}

float
min(float x, float y)
{
    return (x < y ? x : y);
}

float
min(float * a, int size)
{
#ifdef USE_VDSP
    float r;
    vDSP_minv(a, 1, &r, size);
    return r;
#else
    int mi = 0;
    for (int i=0; i<size; i++)
        if (a[i] < a[mi])
            mi = i;
    return a[mi];
#endif
}

float
min(float ** a, int sizex, int sizey)
{
    return min(*a, sizex*sizey);
}

float *
min(float * r, float * a, int size)
{
#ifdef USE_VDSP
    vDSP_vmin(a, 1, r, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        if (a[i] > r[i])
            r[i] = a[i];
#endif
    return r;
}

float **
min(float ** r, float ** a, int sizex, int sizey)
{
    min(*r, *a, sizex*sizey);
    return r;
}

float *
min(float * r, float * a, float * b, int size)
{
#ifdef USE_VDSP
    vDSP_vmin(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        if (a[i] > b[i])
            r[i] = a[i];
        else
            r[i] = b[i];
#endif
    return r;
}

float **
min(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    min(*r, *a, *b, sizex*sizey);
    return r;
}

int
arg_min(float * a, int size)
{
    if (size < 1)
        return -1;
    int mi = 0;
    for (int i=0; i<size; i++)
        if (a[i] < a[mi])
            mi = i;
    return mi;
}

void
arg_min(int & x, int & y, float ** a, int sizex, int sizey)
{
    float t;
    float min = a[x][y];
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            if ((t = a[j][i]) < min)
            {
                x = i;
                y = j;
                min = t;
            }
}

// max
int
max(int x, int y)
{
    return (x > y ? x : y);
}

float
max(float x, float y)
{
    return (x > y ? x : y);
}

float
max(float * a, int size)
{
#ifdef USE_VDSP
    float r;
    vDSP_maxv(a, 1, &r, size);
    return r;
#else
    int mi = 0;
    for (int i=0; i<size; i++)
        if (a[i] > a[mi])
            mi = i;
    return a[mi];
#endif
}

float
max(float ** a, int sizex, int sizey)
{
    return max(*a, sizex*sizey);
}

float *
max(float * r, float * a, int size)
{
#ifdef USE_VDSP
    vDSP_vmax(a, 1, r, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        if (a[i] < r[i])
            r[i] = a[i];
#endif
    return r;
}

float **
max(float ** r, float ** a, int sizex, int sizey)
{
    max(*r, *a, sizex*sizey);
    return r;
}

float *
max(float * r, float * a, float * b, int size)
{
#ifdef USE_VDSP
    vDSP_vmax(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        if (a[i] < b[i])
            r[i] = a[i];
        else
            r[i] = b[i];
#endif
    return r;
}

float **
max(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    max(*r, *a, *b, sizex*sizey);
    return r;
}

int
arg_max(float * a, int size)
{
    if (size < 1)
        return -1;
    float m = a[0];
    float t;
    int mi = 0;
    for (int i=1; i<size; i++)
        if ((t = a[i]) > m)
        {
            mi = i;
            m = t;
        }
    return mi;
}

void
arg_max(int & x, int & y, float ** a, int sizex, int sizey)
{
    float t;
    float max = a[x][y];
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            if ((t = a[j][i]) > max)
            {
                x = i;
                y = j;
                max = t;
            }
}

// minmax
float *
minmax(float & min, float & max, float * a, int size)
{
    min = *a;
    max = *a;
    float t;
    for (int i=0; i<size; i++)
        if ((t = *a++) > max)
            max = t;
        else if (t < min)
            min = t;
    return a;
}

float **
minmax(float & min, float & max, float ** a, int sizex, int sizey)
{
    minmax(min, max, *a, sizex*sizey);
    return a;
}

// mean
float
mean(float * a, int size)
{
#ifdef USE_VDSP
    float r;
    vDSP_meanv(a, 1, &r, size);
    return r;
#else
    float s = 0;
    for (int i=0; i<size; i++)
        s += a[i];
    if (s == 0)
        return 0;
    else
        return s/float(size);
#endif
}

float
mean(float ** a, int sizex, int sizey)
{
    return mean(*a, sizex*sizey);
}

// clip
float
clip(float x, float low, float high)
{
    if (x < low)
        return low;
    else if (x > high)
        return high;
    else
        return x;
}

float *
clip(float * a, float low, float high, int size)
{
#ifdef USE_VDSP
    vDSP_vclip(a, 1, &low, &high, a, 1, size);
#else
    for (int i=0; i<size; i++)
        if (a[i] <low )
            a[i] = low;
        else if (a[i] > high)
            a[i] = high;
#endif
    return a;
}

float **
clip(float ** a, float low, float high, int sizex, int sizey)
{
    clip(*a, low, high, sizex*sizey);
    return a;
}

float *
clip(float * r, float * a, float low, float high, int size)
{
#ifdef USE_VDSP
    vDSP_vclip(a, 1, &low, &high, r, 1, size);
#else
    for (int i=0; i<size; i++)
        if (a[i] <low )
            r[i] = low;
        else if (a[i] > high)
            r[i] = high;
        else
            r[i] = a[i];
#endif
    return r;
}

float **
clip(float ** r, float ** a, float low, float high, int sizex, int sizey)
{
    clip(*r, *a, low, high, sizex*sizey);
    return r;
}

// add
float
add(float * a, int size)	// sum a
{
    float r = 0;
#ifdef USE_VDSP
    vDSP_sve(a, 1, &r, size);
#else
    for (int i=0; i<size; i++)
        r += a[i];
#endif
    return r;
}

float
add(float ** a, int sizex, int sizey)	// sum a
{
    return add(*a, sizex*sizey);
}

float *
add(float * r, float alpha, int size)	// r = r + alpha
{
#ifdef USE_VDSP
    vDSP_vsadd(r, 1, &alpha, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] += alpha;
#endif
    return r;
}

float **
add(float ** r, float alpha, int sizex, int sizey)	// r = r + alpha
{
    add(*r, alpha, sizex*sizey);
    return r;
}

float *
add(float * r, float * a, int size)	// r = r + a
{
    return add(r, r, a, size);
}

float **
add(float ** r, float ** a, int sizex, int sizey)	// r = r + a
{
    add(*r, *r, *a, sizex*sizey);
    return r;
}

float *
add(float * r, float * a, float * b, int size)	// r = a + b
{
#ifdef USE_VDSP
    vDSP_vadd(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = a[i] + b[i];
#endif
    return r;
}

float **
add(float ** r, float ** a, float ** b, int sizex, int sizey)	// r = a + b
{
    add(*r, *a, *b, sizex*sizey);
    return r;
}

float *
add(float * r, float alpha, float * a, int size)	// r = r + alpha * a
{
#ifdef USE_VDSP
    vDSP_vsma(a, 1, &alpha, r, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = r[i] + alpha * a[i];
#endif
    return r;
}

float **
add(float ** r, float alpha, float ** a, int sizex, int sizey)	// r = r + alpha * a
{
    add(*r, alpha, *a, sizex*sizey);
    return r;
}

float *
add(float * r, float alpha, float * a, float beta, float * b, int size)	// r = alpha * a + beta * b
{
#ifdef USE_VDSP
    vDSP_vsmul(a, 1, & alpha, r, 1, size);
    vDSP_vsma(b, 1, &beta, r, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        (*r++) = alpha*(*a++) + beta*(*b++);
#endif
    return r;
}

float **
add(float ** r, float alpha, float ** a, float beta, float ** b, int sizex, int sizey)	// r = alpha * a + beta * b
{
    add(*r, alpha, *a, beta, *b, sizex*sizey);
    return r;
}

float *
add(float * r, float alpha, float * a, float beta, int size)	// r = alpha * a + beta
{
#ifdef USE_VDSP
    vDSP_vsmsa(a, 1, &alpha, &beta, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = alpha * a[i] + beta;
#endif
    return r;
}

float **
add(float ** r, float alpha, float ** a, float beta, int sizex, int sizey)	// r = alpha * a + beta
{
    add(*r, alpha, *a, beta, sizex*sizey);
    return r;
}

// subtract

float *
subtract(float * r, float alpha, int size)	// r = r - alpha
{
    return add(r, -alpha, size);
}

float **
subtract(float ** r, float alpha, int sizex, int sizey)	// r = r - alpha
{
    return add(r, -alpha, sizex, sizey);
}

float *
subtract(float alpha, float * r, int size)	// r = alpha - r
{
    for (int i=0; i<size; i++)
        r[i] = alpha - r[i];
    return r;
}

float **
subtract(float alpha, float ** r, int sizex, int sizey)	// r = alpha - r
{
    subtract(alpha, *r, sizex*sizey);
    return r;
}

float *
subtract(float * r, float * a, int size)
{
    return subtract(r, r, a, size);
}

float **
subtract(float ** r, float ** a, int sizex, int sizey)
{
    subtract(*r, *a, sizex*sizey);
    return r;
}


float *
subtract(float * r, float * a, float * b, int size)
{
#ifdef USE_VDSP
    vDSP_vsub(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = a[i] - b[i];
#endif
    return r;
}


float **
subtract(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    subtract(*r, *a, *b, sizex*sizey);
    return r;
}

// multiply
float *
multiply(float * a, float alpha, int size)
{
#ifdef USE_BLAS
    cblas_sscal(size, alpha, a, 1);
#elif defined USE_VDSP
    vDSP_vsmul(a, 1, &alpha, a, 1, size);
#else
    for (int i=0; i<size; i++)
        a[i] *= alpha;
#endif
    return a;
}

float **
multiply(float ** a, float alpha, int sizex, int sizey)
{
    multiply(*a, alpha, sizex*sizey);
    return a;
}

float *
multiply(float * r, float * a, float alpha, int size)
{
#ifdef USE_BLAS
    reset_array(r, size);
    cblas_saxpy(size, alpha, a, 1, r, 1);
#elif defined USE_VDSP
    vDSP_vsmul(a, 1, &alpha, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = alpha * a[i];
#endif
    return r;
}

float **
multiply(float ** r, float ** a, float alpha, int sizex, int sizey)
{
    multiply(*r, *a, alpha, sizex*sizey);
    return r;
}

float *
multiply(float * r, float * a, int size)
{
    return multiply(r, r, a, size);
}

float **
multiply(float ** r, float ** a, int sizex, int sizey)
{
    multiply(*r, *a, sizex*sizey);
    return r;
}

float *
multiply(float * r, float * a, float * b, int size)
{
#ifdef USE_VDSP
    vDSP_vmul(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = a[i] * b[i];
#endif
    return r;
}

float **
multiply(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    multiply(*r, *a, *b, sizex*sizey);
    return r;
}

// matrix * vector
float *
multiply(float * r, float ** a, float * b, int sizex, int sizey)
{
#ifdef USE_BLAS
    cblas_sgemv(CblasRowMajor, CblasNoTrans, sizey, sizex, 1.0, *a, sizex, b, 1, 0.0, r, 1);
#else
    for (int j=0; j<sizey; j++)
    {
        float s = 0;
        float * aj = a[j];
        float * bi = b;
        for (int i=0; i<sizex; i++)
            s +=  (*aj++) * (*bi++);
        (*r++) = s;
    }
#endif
    return r;
}

// multiply - matrix * matrix
float **
multiply(float ** r, float ** a, float ** b, int sizex, int sizey, int n)
{
#ifdef USE_BLAS
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                sizey, sizex, n, 1.0, *a, n, *b, sizex, 0.0, *r, sizex);
#else
    float s;
    int n1 = n-1;
    for (int j=0; j<sizey; j++)
    {
        for (int i=0; i<sizex; i++)
        {
            s = 0;
            float * aj = a[j];
            float * bki = &b[0][i];
            for (int k=n1; k; k--)
            {
                s += (*aj++) * (*bki);
                bki += sizex;
            }
            r[j][i] = s;
        }
    }
#endif
    return r;
}

// dot - scalar product
float
dot(float * a, float * b, int size)
{
#ifdef USE_BLAS
    return cblas_sdot(size, a, 1, b, 1);
#elif defined USE_VDSP
    float r = 0;
    vDSP_dotpr(a, 1, b, 1, &r, size);
    return r;
#else
    float r = 0;
    for (int i=0; i<size; i++)
        r += a[i] * b[i];
    return r;
#endif
}

float
dot(float ** a, float ** b, int sizex, int sizey)
{
    return dot(*a, *b, sizex*sizey);
}

// outer - tensor product
float **
outer(float ** r, float * a, float * b, int sizex, int sizey)
{
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            r[j][i] = a[i] * b[j];
    return r;
}

// divide
float *
divide(float * r, float * a, int size)
{
    return divide(r, r, a, size);
}

float **
divide(float ** r, float ** a, int sizex, int sizey)
{
    divide(*r, *a, sizex*sizey);
    return r;
}

float *
divide(float * r, float * a, float * b, int size)
{
#ifdef USE_VDSP
    vDSP_vdiv(a, 1, b, 1, r, 1, size);
#else
    for (int i=0; i<size; i++)
        r[i] = a[i] / b[i];
#endif
    return r;
}

float **
divide(float ** r, float ** a, float ** b, int sizex, int sizey)
{
    divide(*r, *a, *b, sizex*sizey);
    return r;
}

// conversion
int
lround(float x)
{
#ifdef WINDOWS32
    if ( x >= (int)x + 0.5f )
        return (int)x + 1;
    else
        return (int)x;
#else
    return ::lroundf(x);
#endif
}

void
float_to_byte(unsigned char * r, float * a, float min, float max, int size)
{
#ifdef USE_VIMAGE
    struct vImage_Buffer src =
        {
            a, 1, size, size*sizeof(float)
        };
    struct vImage_Buffer dest =
        {
            r, 1, size, size*sizeof(float)
        };
    vImage_Error err = vImageConvert_PlanarFtoPlanar8 (&src, &dest, max, min, 0);
    if (err < 0) printf("IKARO_Math::float_to_byte vImage_Error = %ld\n", err);
#else
    for (int i=0; i<size; i++)
        if (a[i] < min)
            r[i] = 0;
        else if (a[i] > max)
            r[i] = 255;
        else
            r[i] = int(255.0*(a[i]-min)/(max-min));
#endif
}

void
byte_to_float(float * r, unsigned char * a, float min, float max, int size)
{
    for (int i=0; i<size; i++)
        r[i] = ((max-min)/255.0) * float(a[i]) + min;
}

int
string_to_int(char * s, int d)
{
    if (s == NULL)
        return d;
    else
        return strtol(s, NULL, 0);	// allows hex and octal formats as well as decimal
}

float
string_to_float(char * s, float d)
{
    if (s == NULL)
        return d;
    else
#ifdef WINDOWS32
        return (float)strtod(s, NULL);
#else
        return strtof(s, NULL);
#endif
}

//  image processing
float **
convolve(float ** result, float ** source, float ** kernel, int rsizex, int rsizey, int ksizex, int ksizey, float bias)
{
#ifdef USE_VIMAGE
    if (ksizex % 2 == 1 && ksizey % 2 == 1)
    {
        struct vImage_Buffer src =
            {
                *source, rsizey+(ksizey-1), rsizex+(ksizex-1), sizeof(float)*(rsizex+(ksizex-1))
            };
        struct vImage_Buffer dest =
            {
                *result, rsizey, rsizex, sizeof(float)*rsizex
            };
        vImage_Error err;
        if (bias != 0)
            err = vImageConvolveWithBias_PlanarF(&src, &dest, NULL, (ksizex-1)/2, (ksizey-1)/2, *kernel, ksizex, ksizey, bias, 0, kvImageTruncateKernel);
        else
            err = vImageConvolve_PlanarF(&src, &dest, NULL, (ksizex-1)/2, (ksizey-1)/2, *kernel, ksizex, ksizey, 0, kvImageTruncateKernel);
        if (err < 0) printf("IKARO_Math::convolve vImage_Error = %ld\n", err);
    }
    else
    {
        // If kernel sizes are not odd use scalar version instead
        for (int j=0; j<rsizey; j++)
            for (int i=0; i<rsizex; i++)
            {
                float s = 0.0;
                for (int l=0; l<ksizey; l++)
                {
                    float * src = &source[j+l][i];
                    float * krn = kernel[l];
                    for (int k=0; k<ksizex; k++)
                        s += krn[k] * (*src++);
                }
                result[j][i] = s + bias;
            }
    }
#else
    for (int j=0; j<rsizey; j++)
        for (int i=0; i<rsizex; i++)
        {
            float s = 0.0;
            for (int l=0; l<ksizey; l++)
            {
                float * src = &source[j+l][i];
                float * krn = kernel[l];
                for (int k=0; k<ksizex; k++)
                    s += krn[k] * (*src++);
            }
            result[j][i] = s + bias;
        }
#endif
    return result;
}

/*
    Reasonably fast implementation of the box filter that sums all the
    matrix elements within each box of size [boxsize] x [boxsize].
    r - matrix with the sums; [sizex] x [sizey]
    a - source matrix; [sizex + boxsize - 1] x [sizey + boxsize - 1]
    t - temporary storage; size [sizex] x [sizey + boxsize - 1]; allocated and deallocated if t == NULL
*/
float **
box_filter(float ** r, float ** a, int sizex, int sizey, int boxsize, bool scale, float ** t)
{
    bool owns_temp = false;
    if (t == NULL)
    {
        t = create_matrix(sizex, sizey+boxsize-1);	// sizex-boxsize+1
        owns_temp = true;
    }
    for (int j=0; j<sizey+boxsize-1; j++)
    {
        t[j][0] = 0;
        for (int i=0; i < boxsize; i++)
            t[j][0] += a[j][i];
    }
    for (int j=0; j<sizey+boxsize-1; j++)
    {
        for (int i=1; i < sizex; i++)
            t[j][i] = t[j][i-1] - a[j][i-1] + a[j][i+boxsize-1];
    }
    for (int i=0; i<sizex; i++)
    {
        r[0][i] = 0;
        for (int j=0; j < boxsize; j++)
            r[0][i] += t[j][i];
    }
    for (int j=1; j < sizey; j++)
    {
        float * sj0 = r[j];
        float * sj1 = r[j-1];
        float * tj1 = t[j-1];
        float * tjb = t[j+boxsize-1];
        for (int i=0; i<sizex; i++)
            *sj0++ = (*sj1++) - (*tj1++) + (*tjb++);
    }
    if (owns_temp)
        destroy_matrix(t);
    if (scale)
        multiply(r, 1.0/sqr(boxsize), sizex, sizey);
    return r;
}

// mics functions
void
select_boltzmann(int & x, int & y, float ** m, int sizex, int sizey, float T)	// *** only calculate exp once!!!
{
    double sum = 0;
    double alpha = 1/T;
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            sum += exp(alpha*m[j][i]);
    if (sum == 0)
    {
        x = sizex/2;
        y = sizey/2;
        return;
    }
    double	rs = random(0.0, sum);
    double	ss = 0;
    for (int j=0; j<sizey; j++)
        for (int i=0; i<sizex; i++)
            if ((ss += exp(alpha*m[j][i])) >= rs)
            {
                x = i;
                y = j;
                return;
            }
}

// Simple minded recursive implementation
//
// Bug: It will not follow a gradient along the edges of the matrix/image ***
void
ascend_gradient(int & x, int & y, float ** m, int sizex, int sizey)
{
    if (x == 0)		return;
    if (y == 0)		return;
    if (x == sizex-1)	return;
    if (y == sizey-1)	return;
    int mx = x;
    int my = y;
    for (int i=-1; i<=1; i++)
        for (int j=-1; j<=1; j++)
            if (m[y+j][x+i] > m[my][mx])
            {
                mx = x+i;
                my = y+j;
            }
    if (mx != x || my != y)
    {
        x = mx;
        y = my;
        ascend_gradient(x, y, m, sizex, sizey);
    }
}

void
descend_gradient(int & x, int & y, float ** m, int sizex, int sizey)
{
    if (x == 0)		return;
    if (y == 0)		return;
    if (x == sizex-1)	return;
    if (y == sizey-1)	return;
    int mx = x;
    int my = y;
    for (int i=-1; i<=1; i++)
        for (int j=-1; j<=1; j++)
            if (m[y+j][x+i] < m[my][mx])
            {
                mx = x+i;
                my = y+j;
            }
    if (mx != x || my != y)
    {
        x = mx;
        y = my;
        descend_gradient(x, y, m, sizex, sizey);
    }
}
}

