//
//    IKAROS_Socket.cc		Socket utilities for the IKAROS project
//
//    Copyright (C) 2001-2006  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    See http://www.ikaros-project.org/ for more information.
//

#include "IKAROS_System.h"

#ifdef USE_BSD_SOCKET

#include "IKAROS_Socket.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>

#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <netdb.h>


#define BACKLOG		10 		// how many pending connections queue will hold
#define TIMEOUT		1000	// number of attempts to read
#define TIMEOUTDEL	1000	// number of usec to sleep before each read



//
//	SocketData
//

class SocketData
{
public:
    int		sd;

    SocketData() : sd(-1)
    {};
};



//
//	Socket - functions that gets data from a HTTP server
//

Socket::Socket()
{
    data = new SocketData();
}



Socket::~Socket()
{
    if (data->sd != -1) close(data->sd);
    delete data;
}



bool
Socket::SendRequest(char * hostname, int port, char * request)
{
    data->sd = -1;
    struct sockaddr_in localAddr, servAddr;
    struct hostent *h;

    if ((h = gethostbyname(hostname)) == NULL)
        return false;  // unkown host

    servAddr.sin_family = h->h_addrtype;
    memcpy((char *) &servAddr.sin_addr.s_addr, h->h_addr_list[0], h->h_length);
    servAddr.sin_port = htons(port);

    if ((data->sd = socket(AF_INET, SOCK_STREAM, 0)) == 0)
        return false; // cannot open socket

    localAddr.sin_family = AF_INET;
    localAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    localAddr.sin_port = htons(0);

    if (bind(data->sd, (struct sockaddr *) &localAddr, sizeof(localAddr)) < 0)
        return false; // cannot open port

    if (connect(data->sd, (struct sockaddr *) &servAddr, sizeof(servAddr)) <0)
        return false; // cannot connect

    if (write(data->sd, request, strlen(request)) <0)
        return false; // cannot send data

    return data->sd != -1;
}



bool
Socket::Poll()
{
    struct timeval tv;
    fd_set readfds;
    tv.tv_sec = 0;
    tv.tv_usec = 0;

    FD_ZERO(&readfds);
    FD_SET(data->sd, &readfds);

    select(data->sd+1, &readfds, NULL, NULL, &tv);

    if (FD_ISSET(data->sd, &readfds))
        return true;
    else
        return false;
}



int
Socket::ReadData(char * result, int maxlen, bool fill)
{
    if (data->sd == -1)
        return 0;

    const int BUFFER_SIZE = 1024;
    static char buffer[BUFFER_SIZE];

    int dst=0;
    int rc = 0;
    do
    {
        int read_size = (maxlen-dst < BUFFER_SIZE ? maxlen-dst : BUFFER_SIZE);
        rc = read(data->sd, buffer, read_size);
        for (int i=0; i<rc && dst<maxlen; i++)
            result[dst++] = buffer[i];

    }
    while (dst < maxlen && (rc >0 || fill));

    return dst;
}



int
Socket::Get(char * hostname, int port, char * request, char * result, int maxlen)
{
    int rc = 0;
    if (SendRequest(hostname, port, request))
        rc = ReadData(result, maxlen);
    Close();

    return rc;
}



void
Socket::Close()
{
    if (data->sd == -1)
        return;
    close(data->sd);
    data->sd = -1;
}



//
//	ServerSocketData
//

class ServerSocketData
{
public:
    int				portno;
    int 			sockfd;						// Listen on sock_fd,
    int				new_fd;						// New connection on new_fd
    int				block_flags;					// Original flags for blocking I/O
    char			request[1024];
    char			argument[10][256];
    int				argcnt;

    char			filepath[PATH_MAX];

    struct sockaddr_in	my_addr; 						// My address information
    struct sockaddr_in	their_addr;					// Connector's address information
};



//
//	ServerSocket
//

ServerSocket::ServerSocket(int port)
{
    data = new ServerSocketData();

    data->portno = port;
    int yes = 1;
    data->argcnt = 0;
    if ((data->sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
        throw SocketException(strerror(errno), 1);

    data->my_addr.sin_family = AF_INET;                     // host byte order
    data->my_addr.sin_port = htons(port);					// short, network byte order
    data->my_addr.sin_addr.s_addr = htonl(INADDR_ANY);		// automatically fill with my IP
    memset(&(data->my_addr.sin_zero), '\0', 8);				// zero the rest of the struct

    if (setsockopt(data->sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1)	// To avoid (most) socket already in use errors
        throw SocketException(strerror(errno), 2);

    if (bind(data->sockfd, (struct sockaddr *)&(data->my_addr), sizeof(struct sockaddr)) == -1)
        throw SocketException(strerror(errno), 3);			// Socket already used

    if (listen(data->sockfd, BACKLOG) == -1)
        throw SocketException(strerror(errno), 4);

    signal(SIGPIPE, SIG_IGN);							// ignore SIGPIPE; catch errors when send() returns -1 instead
}

ServerSocket::~ServerSocket()
{}

bool
ServerSocket::Poll()
{
    int sin_size = sizeof(struct sockaddr_in);
    fcntl(data->sockfd, F_SETFL, O_NONBLOCK);				// temporariliy make the socket non-blocking
    data->new_fd = accept(data->sockfd, (struct sockaddr *)&(data->their_addr), (socklen_t*) &sin_size);
    fcntl(data->sockfd, F_SETFL, 0);							// DON'T make the socket non-blocking
    return (data->new_fd != -1);
}



int
ServerSocket::Read(char * buffer, int maxSize)
{
    if (data->new_fd == -1)
        return 0;

    int n = 0;
    bzero(buffer, maxSize);

    for (int t = 0; t < TIMEOUT; t++)
    {
        if ((n = recv(data->new_fd, buffer, maxSize-1, 0)) >= 0)
            return n;

        usleep(TIMEOUTDEL);
    }

    return n;
}



bool
ServerSocket::GetRequest()
{
    if (!Poll())
        return false;

    if (Read(data->request, 1024) == 0)
        return false;

    // Change scanning ***
    data->argcnt = sscanf(data->request, "%s%s%s%s%s%s%s%s%s%s",
                          data->argument[0], data->argument[1], data->argument[2], data->argument[3],
                          data->argument[4], data->argument[5], data->argument[6], data->argument[7],
                          data->argument[8], data->argument[9]);

    // Remove timestamp in argument 1

    for (int i=0; i<256; i++)
        if (data->argument[1][i] == '#')
            data->argument[1][i] = 0;

    for (int i=0; i<256; i++)
        if (data->argument[1][i] == '?')
            data->argument[1][i] = 0;

//    printf("REQUEST: %s %s %s %s\n", data->argument[0], data->argument[1], data->argument[2], data->argument[3]);

    return true;
}



bool
ServerSocket::TestArgument(int i, char * s)
{
    if (i >= 0 && i<data->argcnt)
        return !strcmp(data->argument[i], s);
    else
        return false;
}



bool
ServerSocket::TestArgumentExtension(int i, char * s)	// Very unsafe!
{
    if (i < 0 && i>=data->argcnt)
        return false;

    int p = strlen(s) - 1;
    int q = strlen(data->argument[i]) - 1;

    while (p >= 0)
    {
        if (data->argument[i][q] == '.')
            return true;
        if (s[p] == '.')
            return true;
        if (data->argument[i][q] != s[p])
            return false;
        p--;
        q--;
    }

    return false;
}



bool
ServerSocket::TestPath(int i, char * s)
{
    if (i < 0 || i>= data->argcnt)
        return false;

    int len = strlen(s);
    for (int j=1; j<len; j++)
        if (data->argument[i][j] == '/')
            return true;
        else if (s[j] != data->argument[i][j])
            return false;

    return true;
}



char *
ServerSocket::GetArgument(int i)
{
    if (i >= 0 && i<data->argcnt)
        return data->argument[i];
    else
        return NULL;
}



// const char month[] = "JanFebMarAprMayJunJulAugSepOctNovDec ";
// const char day[] = "SunMonTueWedThuFriSat";

bool
ServerSocket::SendHTTPHeader(char * type, long size)
{
    Send("HTTP/1.1 200 OK\n");

    if (type != NULL)
        Send("Content-Type: %s\n", type);

    if (size >= 0)
        Send("Content-Length: %ld\n", size);

    Send("\n");

    return true;
}



bool
ServerSocket::SendData(char * buffer, long size)
{
    if (data->new_fd == -1) return false;	// Connection closed - ignore send

    long total = 0;
    long bytesleft = size;
    long n = 0;

    while (total < size)
    {
        n = send(data->new_fd, buffer+total, bytesleft, 0);
        if (n == -1) break;
        total += n;
        bytesleft -= n;
    }

    if (n == -1 || bytesleft > 0) // We failed to send all data
    {
    //    perror("ServerSocket::SendData Failed");
        Close();	// Close the socket and ignore further data
        return false;
    }

    return true;
}



bool
ServerSocket::Send(char * format, ...)
{
    if (data->new_fd == -1) return false;	// Connection closed - ignore send

    char buffer[1024];
    va_list 	args;
    va_start(args, format);

    if (vsnprintf(buffer, 1024, format, args) == -1)
    {
        va_end(args);
        return false;
    }
    va_end(args);

    bool r = SendData(buffer, strlen(buffer));

    return r;
}



bool
ServerSocket::Send(float * data, int sizex, int sizey)
{
    float sx = float(sizex);
    float sy = float(sizey);

    if (!SendData((char *)(&sx), sizeof(float)))
        return false;

    if (!SendData((char *)(&sy), sizeof(float)))
        return false;

    if (!SendData((char *)data, sizex*sizey*sizeof(float)))
        return false;

    return true;
}



bool
ServerSocket::SendFile(char * path, char * filename)
{
    if (filename == NULL) return false;

    if (path != NULL)
        strncpy(data->filepath, path, PATH_MAX);
    else
        data->filepath[0] = 0;

    FILE * f = fopen(strncat(data->filepath, filename, PATH_MAX), "r+b");

    if (f == NULL) return false;

    fseek(f, 0, SEEK_END);
    int len  = ftell(f);
    fseek(f, 0, SEEK_SET);

    char * s = new char[len];
    fread(s, len, 1, f);
    SendData(s, len);
    fclose(f);
    delete s;

    return true;
}



bool
ServerSocket::Close()
{
    if (data->new_fd == -1) return true;	// Connection already closed

    if (close(data->new_fd) == -1)
    {
        errcode = errno;
        return false;
    }
    data->new_fd = -1;

    return true;
}



int
ServerSocket::Port()
{
    return data->portno;
}



char *
ServerSocket::Host()
{
    static char hostname[256];
    if (gethostname(hostname, 256) == 0)	// everything ok
        return hostname;
    else
        return "unkown-host";
}

#endif
