//
//	IKAROS_Utils.cc		Various utilities for the IKAROS project
//
//    Copyright (C) 2001-2002  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	Created: July 17 2001
//
#include "IKAROS_Utils.h"
#include "IKAROS_System.h"

#ifdef WINDOWS
#include <direct.h>
#else
#include <unistd.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#if defined USE_VDSP
#include <Accelerate/Accelerate.h>
#endif




char *
create_string(char * c)
{
    if (c)
    {
        char * p = strcpy(new char [strlen(c)+1], c);
        return p;
    }
    else
        return NULL;
}

char *
create_string_head(char * c, int n)
{
    if (c)
    {
        int l = strlen(c);
    
        if(n > l)
            n = l;

        char * p = strncpy(new char [n+1], c, n);
        p[n] = 0;
        return p;
    }
    else
        return NULL;
}

char *
create_string_tail(char * c, int i)
{
    if (c)
    {
        int l = strlen(c);
    
        if(i > l)
            return NULL;

        char * p = strncpy(new char [l-i+1], &c[i], l-i);
        p[l-i] = 0;
        return p;
    }
    else
        return NULL;
}

char *
create_formatted_string(char *format, ...)
{
    const int max_string = 1024;
    static char temp[max_string];
    va_list args;
    va_start(args, format);
    vsnprintf(temp, max_string, format, args);
    va_end(args);
    int length = strlen(temp)+1;
    char * p = strncpy(new char [length], temp, length);
//	printf("Create formatted string: %s at %p; len = %ld\n", p, p, strlen(p));
    return p;
}

void
destroy_string(char * c)
{
    delete [] c;
}

bool
equal_strings(const char * a, const char * b)
{
    if (a == b)		// Identical string or both NULL
        return true;
    if (a == NULL || b == NULL) // only one is NULL, they can't be equal
        return false;
    return !strcmp(a, b);	// Do the usual comparison
}

bool
is_path(char * p)
{
    if(!p)
        return false;
        
    for(unsigned int i=0; i<strlen(p); i++)
        if(p[i] == '/')
            return true;
        else if (p[i] == '\\')
            return true;
        else if (p[i] == ':')
            return true;
    return false;
}

bool
is_absolute_path(char * p)
{
    if(!p)
        return false;
    if(p[0] == '/')
        return true;
    if(p[1] == ':')
        return true;
    return false;
}

void
print_array(char * name, float * a, int size)
{
    printf("%s[%d] = ", name, size);
    for (int i=0; i<size; i++)
        printf("%7.2f", a[i]);
    printf("\n");
}

void
print_matrix(char * name, float ** m, int sizex, int sizey)
{
    printf("%s[%d][%d] = \n", name, sizey, sizex);
    for (int j=0; j<sizey; j++)
    {
        for (int i=0; i < sizex; i++)
            printf("%8.2f", m[j][i]);
        printf("\n");
    }
    printf("\n");
}

float *
create_array(int size)
{
    return (float *)calloc(size, sizeof(float));
}

float **
create_matrix(int sizex, int sizey)
{
    float * a = (float *)calloc(sizex*sizey, sizeof(float));
    float ** b = (float **)malloc(sizey*sizeof(float *));
    for (int j=0; j<sizey; j++)
        b[j] = &a[j*sizex];
    return b;
}

float *
reset_array(float * a, int size)
{
#ifdef USE_VDSP
    vDSP_vclr(a, 1, size);
#else
    for (int i=0; i<size; i++)
        a[i] = 0.0;
#endif
    return a;
}

float **
reset_matrix(float ** m, int sizex, int sizey)
{
    reset_array(m[0], sizex*sizey);
    return m;
}

float *
set_array(float * a, float v, int size)
{
    for (int i=0; i<size; i++)
        a[i] = v;
    return a;
}

float **
set_matrix(float ** m, float v, int sizex, int sizey)
{
    set_array(m[0], v, sizex*sizey);
    return m;
}

void
destroy_array(float * a)
{
    free(a);
}

void
destroy_matrix(float ** a)
{
    if (!a) return;
    free(*a);
    free(a);
}


float * copy_array(float * r, float * a, int size)
{
    return (float *)memcpy(r, a, size*sizeof(float));
}

float ** copy_matrix(float ** r, float ** a, int sizex, int sizey)
{
    memcpy(r[0], a[0], sizex*sizey*sizeof(float));
    return r;
}


// Options
Options::Options(int argc, char *argv[])
{
    file_path = NULL;
    file_name = NULL;
    char * file_arg = NULL;
    getcwd(working_dir, 1024);

    for (int i=0; i<255; i++)
    {
        option[i] = false;
        argument[i] = NULL;
    }
    for (int i=1; i<argc; i++)
        if (argv[i][0] == '-')
        {
            char  o = argv[i][1];
            if (o != 0 && argv[i][2] != 0)
                SetOption(o, create_string(&argv[i][2]));
            else
                SetOption(o);
        }
        else if (file_arg == NULL)
        {
            file_arg = argv[i];
        }
        
    // Compute binary directory (if possible)

    binary_dir = NULL;
#ifdef WINDOWS
    if(is_path(argv[0]))
#endif
    {
        char * bin_path;
        if(is_absolute_path(argv[0]))
             bin_path = create_string(argv[0]);
        else
#ifdef WINDOWS
             bin_path = create_formatted_string("%s\\%s", working_dir, argv[0]);
#else
             bin_path = create_formatted_string("%s/%s", working_dir, argv[0]);
#endif
        int i;
        for(i=strlen(bin_path)-1; i>0 && bin_path[i] != '/' && bin_path[i] !='\\'; i--)  ; 
        binary_dir = create_string_head(bin_path, i+1);
        destroy_string(bin_path);
    }
    
    // Compute file directory and name
    
    if(file_arg != NULL)
    {
        if(is_absolute_path(file_arg))
            file_path = create_string(file_arg);
        else
            file_path = create_formatted_string("%s/%s", working_dir, file_arg);
            
        int i;
        for(i=strlen(file_path)-1; i>0 && file_path[i] != '/' && file_path[i] !='\\'; i--)  ; 
        file_dir = create_string_head(file_path, i+1);
        file_name =  create_string_tail(file_path, i+1);
    }
}

Options::~Options()
{
    for (int i=0; i<255; i++)
        delete argument[i];
}

void
Options::SetOption(char c, char * arg)
{
    option[int(c)] = true;
    argument[int(c)] = arg;
}

bool
Options::GetOption(char c)
{
    return option[int(c)];
}

char *
Options::GetArgument(char c)
{
    return argument[int(c)];
}

char *  
Options::GetWorkingDirectory()
{
    return working_dir;
}

char *
Options::GetBinaryDirectory()
{
    return binary_dir;
}

char *
Options::GetFilePath()
{
    return file_path;
}

char *
Options::GetFileDirectory()
{
    return file_dir;
}

char *
Options::GetFileName()
{
    return file_name;
}

void
Options::Print()
{
    printf("Options:\n");
    for (int i=0; i<255; i++)
        if (option[i])
        {
            printf("-%c", i);
            if (argument[i])
                printf(" %s", argument[i]);
            printf("\n");
        }
    if (file_name)
    {
        printf("File Dir: %s\n", file_dir);
        printf("File Name: %s\n", file_name);
    }
}


