//
//    main.cc		The main code for the IKAROS project
//
//    Copyright (C) 2001-2007  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    See http://www.ikaros-project.org/ for more information.
//
//	Created: August 8, 2001
//
//	The main function in IKAROS does two things:
//
//	1. It installs all modules and the kernel
//	2. It calls the kernel to start execution
//


#include <stdio.h>
#include <new>
#include <stdlib.h>

// Kernel

#include "IKAROS.h"
#include "WebUI/WebUI.h"

//
// Modules
//

#include "Modules/Modules.h"
#include "UserModules/UserModules.h"



void
PrintInfo()
{
    printf("Usage:\n");
    printf("\n");
    printf("\tIKAROS [-w#][-W#][-p][-t][-T][-r#][-v][-q][-x][-X][-m][-l][-i][-a][file]\n\n");
    printf("\t-w#   WebUI\n");
    printf("\t-W#   WebUI: debug mode, list requests\n");
    printf("\t-p    profile\n");
    printf("\t-t    threaded\n");
    printf("\t-T    list thread allocation\n");
    printf("\t-r#   real time mode; # is time per tick in milliseconds\n");
    printf("\t-s#   stop after # ticks\n");
    printf("\t-v    verbose mode\n");
    printf("\t-q    quiet mode\n");
    printf("\t-x    list xml after paring/combination\n");
    printf("\t-X    debug xml parser\n");
    printf("\t-m    list modules and connections\n");
    printf("\t-c    list installed classes\n");
    printf("\t-l    list scheduling\n");
    printf("\t-i    list installed functionality sockets, timer etc, type of target system\n");
    printf("\t-a    list all; implies -m -l and -T (if -t is set)\n");
}



int
main(int argc, char *argv[])
{

    if (argc == 1)
    {
        PrintInfo();
        return 0;
    }

    Options * options = new Options(argc, argv);

    if (options->GetOption('v'))
        options->Print();

    // Create and Init kernel

    Kernel	k(options);

    try
    {
        k.ListInfo();

        InitModules(k);
        InitUserModules(k);

        k.Init();

        k.ListClasses();
        k.ListModulesAndConnections();
        k.ListScheduling();
        k.ListThreads();

        // Select UI

        if (!k.Terminate() && options->GetOption('w') || options->GetOption('W'))
        {
#ifdef USE_SOCKET
            WebUI webUI(&k);
            webUI.Run();
#else
            printf("IKAROS was compiled without support for sockets and WebUI\n");
#endif
        }
        else
        {
            k.Run();
        }

        k.PrintTiming();
        k.ListProfiling();

    }

    // Catch Exceptions and Terminate Execution

    catch (std::bad_alloc&)
    {
        k.Notify(msg_exception, "Could not allocate memory. Program terminates.\n");
        return 1;	// MEMORY ERROR
    }
    /*
    	catch(XMLError ex)
    	{
    		k.Notify(msg_exception, "%s at line %d. Program terminates (%d).\n", ex.string, ex.line, ex.internal_reference);
    		return 2;	// XML ERROR
    	}
    */
#ifdef USE_SOCKET
    catch (SocketException ex)
    {
        k.Notify(msg_exception, "Socket(%d): %s\n", ex.internal_reference, ex.string);
        return 3;	// SOCKET ERROR
    }
#endif
    catch (int i)
    {
        //	k.Init();
        k.Notify(msg_exception, "%d. Program terminates.\n", i);
        return i;	// OTHER ERROR
    }
    catch (...)
    {
        k.Init();
        k.Notify(msg_exception, "Undefined exception. Program terminates.\n");
        return -1;	// UNDEFINED ERROR
    }

    return 0;
}


