//
//	GridWorld.cc		This file is a part of the IKAROS project
// 					Implements a simple grid world with obstacles and rewards
//
//    Copyright (C) 2003 Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	Created: 2003-07-14
//
//	surrounding: variable size + rotation, direction in image

#include "GridWorld.h"

#include "ctype.h"


using namespace ikaros;

Module * GridWorld::Create(char * name, Parameter * p)
{
    return new GridWorld(name, p);
}



GridWorld::GridWorld(char * name, Parameter * p):
        Module(p)
{
    move		=	NULL;
    location	=	NULL;
    coordinate	=	NULL;
    obstacles	=	NULL;
    goals		=	NULL;
    reward		=	NULL;
    image		=	NULL;

    AddInput("MOVE");

    AddOutput("REWARD", 1);
    AddOutput("COORDINATE", 2);

    AddOutput("OBSTACLES", 3, 3);
    AddOutput("GOALS", 3, 3);

    x_start	= GetIntValue(p, "x_start", 1);
    y_start	= GetIntValue(p, "y_start", 1);

    mode	= GetIntValueFromList(p, "mode", "sum/max/relative_sum/relative_max");

    size_x	= GetIntValue("sizex", 1);
    size_y	= GetIntValue("sizey", 1);
    world	= GetMatrix("grid", size_x, size_y);

    AddOutput("IMAGE", size_x, size_y);
    AddOutput("LOCATION", size_x, size_y);
}



void
GridWorld::ClearAgent(int x, int y)
{
    if (world[y][x] > 0)
        image[y][x] = 3.0;
    else
        image[y][x] = 0.0;
}



void
GridWorld::DrawAgent(int x, int y)
{
    image[y][x] = 2.0;
}



void
GridWorld::Init()
{
    move		=	GetInputArray("MOVE");
    location	=	GetOutputMatrix("LOCATION");
    coordinate	=	GetOutputArray("COORDINATE");

    obstacles	=	GetOutputMatrix("OBSTACLES");
    goals		=	GetOutputMatrix("GOALS");

    reward		=	GetOutputArray("REWARD");
    image		=	GetOutputMatrix("IMAGE");

    // Check inputs here?

    // CHECK INPUT SIZE 3 or 4 !!!!!

    // Draw Image

    for (int i=0; i<size_x; i++)
        for (int j=0; j<size_y; j++)
            if (world[j][i] > 0)
                image[j][i] = 0;

    // Fill in walls

    for (int i=0; i<size_x; i++)
        for (int j=0; j<size_y; j++)
            if (world[j][i] == -1)
                image[j][i] = 0.5;
            else if (world[j][i] > 0)
                image[j][i] = 3;

    DrawAgent(x_start, y_start);
    x = x_start;
    y = y_start;
}



GridWorld::~GridWorld()
{
    destroy_matrix(world);
}


static int dx[4] = { 0, 1, 0, -1 };
static int dy[4] = { -1, 0, 1, 0 };

void
GridWorld::Tick()
{
    ClearAgent(x, y);

    if (reward != NULL && reward[0] > 0)
    {
        reward[0] = 0.0;
        x = x_start;
        y = y_start;
        reward[0] = world[y][x];
        coordinate[0] = x;
        coordinate[1] = y;

        // The rest should be used in ONE place

        reset_matrix(location, size_x, size_y);
        location[y][x] = 1;
        DrawAgent(x, y);

        // Set surrounding

        for (int i=0; i<3; i++)
            for (int j=0; j<3; j++)
                obstacles[j][i] = (world[y+j-1][x+i-1] == -1 ? 1 : 0);

        for (int i=0; i<3; i++)
            for (int j=0; j<3; j++)
                goals[j][i] = (world[y+j-1][x+i-1] > 0 ? world[y+j-1][x+i-1] : 0);

        return;
    }

    if (move == NULL || zero(move, (mode < 2 ? 4 : 3)))
    {
        DrawAgent(x, y);
        return;
    }

    int xr = 0;
    int yr = 0;
    int mv = 0;

    switch (mode)
    {
    case 0: // sum
    {
        float m[4] = {move[0]-move[2], move[1]-move[3], move[2]-move[0], move[3]-move[1]};
        mv = arg_max(m, 4);
        xr = dx[mv];
        yr = dy[mv];
    }	break;

    case 1: // max
    {
        mv = arg_max(move, 4);
        xr = dx[mv];
        yr = dy[mv];
    }	break;

    case 2: // relative sum
    {
        float m[3] = {move[0]-move[1], move[1]-move[0], move[3]};
        mv = arg_max(m, 3);
        if (mv == 2)
        {
            xr = dx[dir];
            yr = dy[dir];
        }
        else if (mv == 1)
            dir = (dir + 1) % 4;
        else if (mv == 1)
            dir = (dir - 1 + 4) % 4;
    }	break;

    case 3: // relative max
    {
        mv = arg_max(move, 3);
        if (mv == 2)
        {
            xr = dx[dir];
            yr = dy[dir];
        }
        else if (mv == 1)
            dir = (dir + 1) % 4;
        else if (mv == 1)
            dir = (dir - 1 + 4) % 4;
    }	break;

    default:
        ;
    }


    // Agent can not occupy the outer border of the grid

    if (x+xr > 0 && x+xr < size_x-1 && y+yr > 0 && y+yr < size_y-1 && world[y+yr][x+xr] != -1)
    {
        x += xr;
        y += yr;
    }

    DrawAgent(x, y);

    reward[0] = world[y][x];
    coordinate[0] = x;
    coordinate[1] = y;
    reset_matrix(location, size_x, size_y);
    location[y][x] = 1.0;

    // Set surrounding

    for (int i=0; i<3; i++)
        for (int j=0; j<3; j++)
            obstacles[j][i] = (world[y+j-1][x+i-1] == -1 ? 1 : 0);

    for (int i=0; i<3; i++)
        for (int j=0; j<3; j++)
            goals[j][i] = (world[y+j-1][x+i-1] > 0 ? world[y+j-1][x+i-1] : 0);
}


