//
//	InputQTAudio.cc		This file is a part of the IKAROS project
//						A module for reading audio from Quicktime files.
//
//    Copyright (C) 2001-2003  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//		CGP; 3 August 2005; Created based on InputQTMovie module.


#include "InputQTAudio.h"
#ifdef USE_QUICKTIME

using namespace qt;

#include "QuickTimeAudioUtils.h"
#include "MTAudioBufferListUtils.h"

#define	BailNULL(n)  if (!n) goto bail;
#define	BailError(n) if (n) goto bail;



typedef NavObjectFilterUPP	QTFrameFileFilterUPP;



//------------------------------------------------------------------------------
//	Open up the Movie file and get a Movie from it.
//------------------------------------------------------------------------------
void
InputQTAudio::OpenMovie(UInt8 * filename)
{
    FSRef	theFSRef;
    FSSpec	theFSSpec;
    short	refnum = 0;
    Boolean  isDirectory;

    OSErr err = noErr;

    err =FSPathMakeRef (filename, &theFSRef, &isDirectory);
    BailError(err);

    err =FSGetCatalogInfo (&theFSRef, kFSCatInfoNone, NULL, NULL, &theFSSpec, NULL);
    BailError(err);

    err = OpenMovieFile(&theFSSpec, &refnum, fsRdPerm);
    BailError(err);

    err = NewMovieFromFile(&movie, refnum, NULL, NULL, newMovieActive, NULL);
    BailError(err);

bail:
    if (refnum)
        CloseMovieFile(refnum);
}



UInt32 InputQTAudio::GetNumSamplesInNextFrame()
{
    TimeRecord timeRecord;
    OSType myTypes[1];
    myTypes[0] = VisualMediaCharacteristic;					// we want video samples
    short interestingTime = nextTimeStep;
    static bool endOfMovie = false;

    /*
    If you keep accumulating time from frame to frame, and then extracting audio samples of that duration, 
    then you could drift by a fraction of a sample for each frame, due to the rounding in the UInt32 cast. 
    It's unlikely to be a big problem, but nevertheless since the audio is a continuous stream of uninterrupted samples, 
    it would be better if you used the extracted audio sample count to determine your position.

    For instance, in your application, use GetNextInterestingTime to determine the input movie time of the start of the next video frame. 
    Then extract enough audio samples to get from where you last extracted to that point. 
    Keep a running count of the total number of audio samples you have extracted and the output movie time equals ((Float64) totalAudioSampleCount / asbd.mSampleRate).

    The number of audio samples to extract for the next frame is thus:
    (UInt32) ((((Float64) *(TimeValue64*)&nextInterestingTimeRecord.value / (Float64) nextInterestingTimeRecord.scale) * asbd.mSampleRate) - totalAudioSampleCount).

    Daniel Steinberg
    QuickTime Engineering
    */

    GetMovieNextInterestingTime(movie, interestingTime, 1, myTypes, movieTime, fixed1, &movieTime, NULL);

    if ((movieTime == -1) && (! endOfMovie)) {
        // GetMovieNextInterestingTime is indicating there are no more visual frames, but this may not be the end of the movie. If we use this, we may
        // not obtain all of the audio in the movie. So, use the movie duration as the last movieTime.
        if (m_movieDuration > movieTime) {
            movieTime = m_movieDuration;
            printf("InputQTAudio::GetNumSamplesInNextFrame: movieTime == -1, and m_movieDuration > movieTime\n");
        }

        endOfMovie = true; // enter into this conditional only once
    }

    if (movieTime != -1)
    {
        int errCode = GetMoviesError();
        if (errCode != noErr)
        {
            Notify(msg_warning, "QuickTime ErrCode = %d (t = %d)\n", errCode, (int) movieTime);
        }

        // Now, set our movie to be positioned at this interesting time, so we can get the TimeRecord
        SetMovieTimeValue(movie, movieTime);

        // Get the TimeRecord
        GetMovieTime(movie, &timeRecord);

//		printf("InputQTAudio::GetNumSamplesInNextFrame: time record: value= %ld, scale= %f\n",  (long)  *((TimeValue64*) &timeRecord.value), (Float64) (timeRecord.scale));

        // Convert to floating-point seconds
        Float64 nextTime = (*((TimeValue64*) &timeRecord.value) / (Float64) timeRecord.scale);

//		printf("InputQTAudio::GetNumSamplesInNextFrame: time at end of next video frame= %f\n", nextTime);

        // This should give me the TOTAL number of audio samples (i.e., samples/sec * seconds = samples) that span from
        // the start of the video until the "nextTime".
        UInt32 totalNumSamples = (UInt32) (m_summaryASBD.mSampleRate * nextTime);

//		printf("InputQTAudio::GetNumSamplesInNextFrame: TOTAL Number of samples= %d\n", (int) totalNumSamples);

        // m_totalAudioSampleCount is the total number of audio samples we have obtained from the movie previously.
        UInt32 result = totalNumSamples - m_totalAudioSampleCount;

//		printf("InputQTAudio::GetNumSamplesInNextFrame: Number of samples to read next= %d\n", (int) result);

        return result;
    }

    Notify(msg_end_of_file);

    return 0;
}

void InputQTAudio::InitAudio()
{
    OSStatus r;

    AudioChannelLayout* outLayout;

//	printf("InputQTAudio::InitAudio: Initializing audio\n");

    r = getDefaultExtractionLayout(movie, NULL, &outLayout, &m_summaryASBD);
    if (r != noErr) {
        printf("InputQTAudio::InitAudio: Error calling getDefaultExtractionLayout\n");
        return;
    }
    /*
    	cout << "mSampleRate= " << m_summaryASBD.mSampleRate << endl;
    	cout << "mFormatID= " << m_summaryASBD.mFormatID << endl;
    	cout << "mFormatFlags= " << m_summaryASBD.mFormatFlags  << endl;
    	cout << "mBytesPerPacket= " << m_summaryASBD.mBytesPerPacket  << endl;
    	cout << "mFramesPerPacket= " << m_summaryASBD.mFramesPerPacket << endl;
    	cout << "mBytesPerFrame= " << m_summaryASBD.mBytesPerFrame << endl;
    	cout << "mChannelsPerFrame= " << m_summaryASBD.mChannelsPerFrame << endl;
    	cout << "mBitsPerChannel= " << m_summaryASBD.mBitsPerChannel << endl;
    	cout << "mReserved= " << m_summaryASBD.mReserved << endl;

    	cout << kAudioFormatLinearPCM << endl <<
    kAudioFormatAC3 << endl <<
    kAudioFormat60958AC3 << endl <<
    kAudioFormatAppleIMA4 << endl <<
    kAudioFormatMPEG4AAC << endl <<
    kAudioFormatMPEG4CELP <<  endl <<
    kAudioFormatMPEG4HVXC << endl <<
    kAudioFormatMPEG4TwinVQ << endl <<
    kAudioFormatTimeCode <<  endl <<
    kAudioFormatMIDIStream << endl <<
    kAudioFormatParameterValueStream << endl;
    */

    // the number of audio channels
//	printf("InputQTAudio::InitAudio: outLayout->mNumberChannelDescriptions= %d\n",  (int) outLayout->mNumberChannelDescriptions);
    m_numChannels = (int) outLayout->mNumberChannelDescriptions;

    r = MovieAudioExtractionBegin(movie, 0, &m_audioRef);
    if (r != noErr) {
        printf("InputQTAudio::InitAudio: Error calling MovieAudioExtractionBegin\n");
        return;
    }

    m_movieDuration = GetMovieDuration(movie);

//	printf("InputQTAudio::InitAudio: Apparent success in initializing audio!\n");

    free(outLayout);
}

InputQTAudio::InputQTAudio(char * name, Parameter * p):
        Module(p)
{
    movie = NULL;
    movieTime = 0;			// set current time value to begining of the Movie
    m_totalAudioSampleCount = 0;

    filename = GetValue(p, "filename");

    if (!filename)
    {
        Notify(msg_fatal_error, "No filename for InputQTAudio.\n");
        return;
    }

    m_requestedSamplesPerBuffer  = GetIntValue(p, "samples_per_buffer", -1);
    if (m_requestedSamplesPerBuffer <= 0) {
        // Just in case the user gives us a strange value
        m_requestedSamplesPerBuffer = -1;
    }

    // enables the user to select only the first n of the audio channels
    m_numChannelsRequested = GetIntValue(p, "num_channels_requested", -1);

//	printf("InputQTAudio::InputQTAudio: m_requestedSamplesPerBuffer= %d\n", m_requestedSamplesPerBuffer); fflush(stdout);

    OSErr result = noErr;

    // This seems to be Mac GUI related-- to make the cursor visible
    InitCursor();

    // Initialize the QuickTime Movie Toolbox
    result = EnterMovies();
    if (result != noErr)
        return;

    // get the movie
    OpenMovie((UInt8 *)filename);
    if (!movie) {
        printf("InputQTAudio: Cannot open movie for filename= %s\n", filename);
        Notify(msg_fatal_error, "InputQTAudio: Cannot open movie for filename= %s\n", filename);
        return;
    }

    // Initialize audio
    InitAudio();

    m_audioData = NULL;

    if (m_requestedSamplesPerBuffer == -1) {
        // User did not request a number of samples per buffer, so synchronize audio buffers with the visual track.

        m_nextNumberOfSamples = GetNumSamplesInNextFrame();

        // Estimate the maximum number of samples as 2 * the number of audio samples associated with the first visual frame.
        // I have no idea if this is the max, but I don't have a better way to estimate the max right now.
        m_maxNumSamples = m_nextNumberOfSamples * 2;
    } else {
        // User did request a number of samples per buffer
        m_nextNumberOfSamples = m_requestedSamplesPerBuffer;
        m_maxNumSamples = m_requestedSamplesPerBuffer;
    }

    if ((m_numChannelsRequested != -1) && (m_numChannelsRequested > 0) && (m_numChannelsRequested < m_numChannels)) {
        AddOutput("CHANNELS", m_numChannelsRequested, m_maxNumSamples);
    } else {
        m_numChannelsRequested = m_numChannels;
        // This output will contain all of the audio channels. Each row will correspond to one audio frame from all channels.
        // Each column will contain data for a single channel. We attempt to allocate the maximum possible number of rows for this output.
        AddOutput("CHANNELS", m_numChannels, m_maxNumSamples);
    }

    // This output will contain a single number on each tick-- the number of rows that have values assigned in the CHANNELS output.
    AddOutput("NUMBER_SAMPLES", 1, 1);

    // Need to allocate the AudioBufferList in order to call MovieAudioExtractionFillBuffer.
    // The documentation indicates that allocation of an AudioBufferList varies depending on whether the audio
    // is interleaved. At this point, I'm assuming that the interleaved option is not specifying how the audio is represented in the file, but
    // rather how the user wants to represent the audio. I.e., I'm assuming that the MovieAudioExtractionFillBuffer call can fill up the
    // AudioBufferList in either a interleaved or non-interleaved fashion.
    m_audioData = MTAudioBufferListNew (m_numChannels, m_maxNumSamples, false);
    if (m_audioData == NULL) {
        Notify(msg_fatal_error, "InputQTAudio::InputQTAudio: Error allocating AudioBufferList\n");
        return;
    }

    // For some ungodly reason MovieAudioExtractionFillBuffer alters the mDataByteSize when called, so this needs to be reset before
    // calling MovieAudioExtractionFillBuffer the second, third etc. times.
    m_channelBufferLength = m_audioData->mBuffers[0].mDataByteSize;
}



InputQTAudio::~InputQTAudio()
{
    if (movie)
        DisposeMovie(movie);

    if (m_audioData != NULL) {
        MTAudioBufferListDispose (m_audioData);
    }
}



Module *
InputQTAudio::Create(char * name, Parameter * p)
{
    return new InputQTAudio(name, p);
}



void
InputQTAudio::Init()
{
    Notify(msg_warning, "InputQTAudio is not thread safe\n");

    channels = GetOutputArray("CHANNELS");
    numberSamples = GetOutputArray("NUMBER_SAMPLES");
}

FILE *fp = NULL;

void
InputQTAudio::Tick()
{
    UInt32 audioFlags;
    OSStatus r;

    r = MovieAudioExtractionFillBuffer( m_audioRef,	// our movie audio extract. session ref
                                        &m_nextNumberOfSamples,	// number of PCM frames to be extracted
                                        m_audioData,		// AudioBufferList
                                        &audioFlags);	// tells us when extraction is complete

    m_totalAudioSampleCount += m_nextNumberOfSamples;

    if (r != noErr) {
        printf("InputQTAudio::Tick: Error calling MovieAudioExtractionFillBuffer\n");
        printf("InputQTAudio::Tick: numSamples= %d, audioFlags= %d\n", (int) m_nextNumberOfSamples, (int) audioFlags);
        return;
    }

    // As noted in the .h file, for some ungodly reason, MovieAudioExtractionFillBuffer alters the mDataByteSize field, so we need to reset it.
    for (unsigned channelIndex=0; channelIndex < m_audioData->mNumberBuffers; channelIndex++) {
        m_audioData->mBuffers[channelIndex].mDataByteSize = m_channelBufferLength;
    }

    printf("InputQTAudio::Tick: numSamples= %d, audioFlags= %d\n", (int) m_nextNumberOfSamples, (int) audioFlags);

    if ((audioFlags == kQTMovieAudioExtractionComplete) && (m_nextNumberOfSamples == 0)) {
        printf("InputQTAudio::Tick: End of file\n");
        Notify(msg_end_of_file);
    }

    // Copy to IKAROS output arrays
    *numberSamples = (Float32) m_nextNumberOfSamples;
    Float32 *output = channels;

    /*
    if (fp == NULL) {
    	fp = fopen("/tmp/debug2.out", "w");
    }
    */

    for (unsigned i=0; i < m_nextNumberOfSamples; i++) {
        //printf("t4- s= %d: ", (int) i);
        for (unsigned channelIndex = 0; channelIndex < (unsigned) m_numChannelsRequested; channelIndex++) {
            *output = ((Float32 *) (m_audioData->mBuffers[channelIndex].mData))[i];
            //printf("c[%d]= %f; ", (int) channelIndex, *output);
            output++;
        }

        //fprintf(fp, "\n"); fflush(fp);
    }

    //printf("\n");

    if (m_requestedSamplesPerBuffer == -1) {
        m_nextNumberOfSamples = GetNumSamplesInNextFrame();

        // Only need to make this check if we are synchronizing with the visual track-- otherwise, we should be reading exactly the number of samples that the user requested
        if (m_nextNumberOfSamples > m_maxNumSamples) {
            printf("InputQTAudio::Tick: maximum number of audio samples exceeded: %d (max: %d)\n", (int) m_nextNumberOfSamples, (int) m_maxNumSamples);
            Notify(msg_fatal_error, "InputQTAudio::Tick: maximum number of audio samples exceeded: %d (max: %d)\n", (int) m_nextNumberOfSamples, (int) m_maxNumSamples);
            return; // shouldn't get to here
        }
    }
}

#endif
