//
//	NetworkCamera.cc		This file is a part of the IKAROS project
//						A module to grab images and control movements of a network camera
//
//    Copyright (C) 2002-2006  Christian Balkenius
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

// NOTE: This code looks as a mess since it glues together code uding three different coding styles:
// libjpeg, Ikaros, and the POSIX socket library.



#include "NetworkCamera.h"

#ifdef USE_LIBJPEG
#ifdef USE_SOCKET

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifdef WINDOWS
#include <direct.h>
#include <winsock.h>
#include <windows.h>
#endif

#ifdef POSIX
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <unistd.h>
#endif

#define MAX_STR_LEN 512



extern "C" {
#include "jpeglib.h"
}

#include <setjmp.h>


void
NetworkCamera::SkipBoundary(char * boundary)
{
    int p = 0, rc;
    char c;

    while (boundary[p] != 0)
    {
        rc = socket->ReadData(&c, 1);
        if (boundary[p] == c)
            p++;
    }

    socket->ReadData(&c, 1); // Skip line break
    socket->ReadData(&c, 1); // Skip line break
}



void
NetworkCamera::ReadLine(char * line)
{
    int p = 0;
    int rc;
    char c;
    do
    {
        rc = socket->ReadData(&c, 1);
        if (rc == 1 && c != '\n' && c!= '\r')
            line[p++] = c;
        else
            line[p] = 0;
    }
    while (c != 13);
    socket->ReadData(&c, 1);  // Skip line break
}



bool
NetworkCamera::ReadBlock()
{
    // Check for new image data on the socket

    if (!socket->Poll())
        return false;

    // Read Header

    char line[256];
    char attribute[256];
    int length = 0;
    do
    {
        ReadLine(line);

        int cnt = sscanf(line, "%[^:]", attribute);

        if (cnt == 1 && !strcmp(attribute, "Content-Length"))
        {
            sscanf(line, "%[^:]:%d", attribute, &length);
//			printf("\tLENGTH = %d\n", length);
        }

        else if (cnt == 1 && !strcmp(attribute, "Delta-time"))
        {
            int delta_time;
            cnt = sscanf(line, "%[^:]:%d", attribute, &delta_time);
            if (cnt > 0)
                timestamp += delta_time;
//			printf("\tTIME = %ld + %d\n", timestamp, delta_time);
        }
    }
    while (strlen(line) > 0);

    // Read Image Data (fill buffer)

    socket->ReadData(buffer, length, true);
    SkipBoundary(boundary);

    return true;
}



void
NetworkCamera::ReadHeader() // This part is specific for the AXIS camera and should be more general
{
    char header[1024];

    int rc = socket->ReadData(header, 100); // we just happen to know that the head size == 100
    header[100] = 0;

    // Read boundary marker

    int b = 0;
    char c;
    do
    {
        rc = socket->ReadData(&c, 1);
        if (rc == 0)
        {
            printf("Error reading boundary marker\n");
            return;
        }
        if (c != '\r' && c != '\n')
            boundary[b++] = c;
    } while (c != '\n');
    boundary[b] = 0;

    // The boundary marker should probably not contain the leading "--" but since it does so we just keep them in the string

    SkipBoundary(boundary);
}



Module *
NetworkCamera::Create(char * name, Parameter * p)
{
    return new NetworkCamera(name, p);
}



NetworkCamera::NetworkCamera(char * name, Parameter * p):
        Module(p)
{
    intensity	=	NULL;
    red		=	NULL;
    green	=	NULL;
    blue		=	NULL;

    timestamp = 0;

    size_x  = GetIntValue(p, "size_x", 352);	// 704x480, 176x144, 352x240
    size_y  = GetIntValue(p, "size_y", 240);

    if (size_x*size_y < 1)
    {
        Notify(msg_fatal_error, "Image size not supplied.\n");
        return;
    }

    host_ip			=	GetValue(p, "host_ip");
    host_port			=	GetIntValue(p, "host_port", 80);
    image_request		=	GetValue(p, "image_request");		// Not used
    fps				=	GetIntValue(p, "fps", 10);
    compression		=	GetIntValue(p, "compression", 50);

    AddOutput("INTENSITY", size_x, size_y);
    AddOutput("RED", size_x, size_y);
    AddOutput("GREEN", size_x, size_y);
    AddOutput("BLUE", size_x, size_y);

    // Create socket

    socket = new Socket();

    // Set up connection and tell camera to stream data

    char msg[1024];
    sprintf(msg, "GET /axis-cgi/mjpg/video.cgi?deltatime=1&showlength=1&resolution=%dx%d&req_fps=%d&compression=%d HTTP/1.1\r\n\r\n", size_x, size_y, fps, compression);

    if (!socket->SendRequest(host_ip, host_port, msg))
    {
        Notify(msg_fatal_error, "Could not connect to network camera at %s:%d.\n", host_ip, host_port);
        return;
    }

    ReadHeader();
}



NetworkCamera::~NetworkCamera()
{

}



void
NetworkCamera::Init()
{
    intensity		= GetOutputArray("INTENSITY");
    red			= GetOutputArray("RED");
    green		= GetOutputArray("GREEN");
    blue			= GetOutputArray("BLUE");
}



//
// IN MEMORY CONVERSION DATA SOURCE MANAGER
//


struct my_error_mgr {
    struct jpeg_error_mgr pub;	// "public" fields
    jmp_buf setjmp_buffer;		// for return to caller
};



typedef struct my_error_mgr * my_error_ptr;



static void
my_error_exit2 (j_common_ptr cinfo)
{
    my_error_ptr myerr = (my_error_ptr) cinfo->err;
    (*cinfo->err->output_message) (cinfo);
    longjmp(myerr->setjmp_buffer, 1);
}



void
init_source2 (j_decompress_ptr cinfo)
{
}



boolean
fill_input_buffer2 (j_decompress_ptr cinfo)
{
    return TRUE;
}



void
skip_input_data2 (j_decompress_ptr cinfo, long num_bytes)
{
    cinfo->src->next_input_byte += (size_t) num_bytes;
}



void
term_source2 (j_decompress_ptr cinfo)
{
}



void
NetworkCamera::DecodeImage(char * source)
{
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    struct jpeg_source_mgr * src;

    JSAMPARRAY buffer;		// Output row buffer
    int row_stride;				// physical row width in output buffer

    // BUILD ERROR HANDLER

    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit2;

    if (setjmp(jerr.setjmp_buffer))
    {
        jpeg_destroy_decompress(&cinfo);
        return;	// REPORT ERRORs
    }

    jpeg_create_decompress(&cinfo);

    // BUILD SOURCE MANAGER

    if (cinfo.src == NULL)	/* first time for this JPEG object? */
        cinfo.src = (struct jpeg_source_mgr *)(*cinfo.mem->alloc_small) ((j_common_ptr)(&cinfo),
                    JPOOL_PERMANENT, sizeof(struct jpeg_source_mgr));

    src = (struct jpeg_source_mgr *) cinfo.src;

    src->init_source = init_source2;
    src->fill_input_buffer = fill_input_buffer2;
    src->skip_input_data = skip_input_data2;
    src->resync_to_restart = jpeg_resync_to_restart;  // Use default method
    src->term_source = term_source2;
    src->bytes_in_buffer = (size_t)(-1);
    src->next_input_byte = (JOCTET *)(source);	// Decode this data!!!

    // START DECODING THE JPEG IMAGE

    (void) jpeg_read_header(&cinfo, TRUE);

    cinfo.scale_denom = 1;

    (void) jpeg_start_decompress(&cinfo);

    row_stride = cinfo.output_width * cinfo.output_components;

//	printf("# width = %d height = %d components = %d\n", cinfo.output_width, cinfo.output_height, cinfo.output_components);

//	*w = cinfo.output_width;
//	*h = cinfo.output_height;

    buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

    if (cinfo.output_components == 1) // Gray Scale Image
    {
        while (cinfo.output_scanline < cinfo.output_height)
        {
            (void) jpeg_read_scanlines(&cinfo, buffer, 1);
            for (int i=0; i<size_x; i++) // CHECK BOUNDS **
            {
                int ix = size_x*(cinfo.output_scanline-1) + i;
                intensity[ix]	= float(buffer[0][i])/255.0;
                red[ix]		= intensity[ix];
                green[ix]		= intensity[ix];
                blue[ix]		= intensity[ix];
            }
        }
    }
    else // RGB Color Image (Optimized version)
    {
        float d255 = 1.0/255.0;
        float d3 = 1.0/3.0;
        while (cinfo.output_scanline < cinfo.output_height)
        {
            (void) jpeg_read_scanlines(&cinfo, buffer, 1);
            unsigned char * z = buffer[0];
            int ix = size_x*(cinfo.output_scanline-1);
            float * r = &red[ix];
            float * g = &green[ix];
            float * b = &blue[ix];
            float * ii = &intensity[ix];
            for (int i=0; i<size_x; i++)
            {
                float s = *r++ = d255*float(*z++);
                s+= *g++ = d255*float(*z++);
                s+= *b++ = d255*float(*z++);
                *ii++ = d3*s;
            }
        }
    }

    (void) jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
}



void
NetworkCamera::Tick()
{
    while (ReadBlock())  ;	// Read all available blocks
    DecodeImage(buffer);	// and decompress the last
}

#endif
#endif

