//
//	ColorTransform.cc		This file is a part of the IKAROS project
//						A module to convert color coordinates
//
//    Copyright (C) 2003  Christian Balkenius, Anders J Johansson
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	Partially based on code by Mark Ruzon from C code by Yossi Rubner, 23 September 1997.
//	Updated for MATLAB 5 28 January 1998. Moved to IKAROS February 2003.


#include "ColorTransform.h"

using namespace ikaros;


// Transformation types

#define 	RGB2LAB		0
#define 	RGB2XYZ		1
#define 	LAB2RGB		2
#define 	XYZ2RGB		3

static inline
float
limit(float x, float a, float b)
{
    return(x < a ? a :  (x > b ? b : x));
}



//inline
static void
transform_rgb2xyz(float red, float green, float blue, float & X, float & Y, float & Z)
{
    X 	= 0.412453*red	+ 0.357581*green	+ 0.180423*blue;
    Y	= 0.212670*red	+ 0.715160*green	+ 0.072169*blue;
    Z	= 0.019334*red	+ 0.119193*green	+ 0.950227*blue;
}



//inline
static void
transform_xyz2rgb(float X, float Y, float Z, float & red, float & green, float & blue)
{
    red 		=   3.240479*X	- 1.537150*Y		- 0.498535*Z;
    green	= -0.969256*X	+ 1.875992*Y		+ 0.041556*Z;
    blue		=  0.055648*X		- 0.204043*Y		+ 1.057311*Z;
}



const float T = 0.008856;



static float
f(float t)
{
    if (t > T)
        return ikaros::pow((t), 1.0/3.0);
    else
        return 7.787*(t) + 16.0/116.0;
}



// transform_rgb2lab transforms from RGB into CIE Lab.
// Based on ITU-R Recommendation  BT.709 using the D65 white point reference.

inline
static void
transform_rgb2lab(float red, float green, float blue, float & L, float & a, float & b)
{
    float X, Y, Z;

    X 	= 0.412453*red	+ 0.357581*green	+ 0.180423*blue;
    Y	= 0.212670*red	+ 0.715160*green	+ 0.072169*blue;
    Z	= 0.019334*red	+ 0.119193*green	+ 0.950227*blue;

    X 	/= 0.950456;
    Y 	/= 1.000000;
    Z 	/= 1.088754;

    if (Y > T)
        L = 116.0*ikaros::pow(Y, 1.0/3.0)-16.0;
    else
        L = 903.3*Y;

    if (L < 0)
        L = 0;

    a = 500.0*( f(X) - f(Y) );
    b = 200.0*( f(Y) - f(Z) );
}



inline
static void
transform_lab2rgb(float L, float a, float b, float & red, float & green, float & blue, float scale = 1.0)
{
    float T1 = 0.008856;
    float T2 = 0.206893;

    // Compute Y

    float	fY = ikaros::pow(((L + 16) / 116), 3.0);
    bool YT = fY > T1;

    if (!YT)
        fY = L / 903.3;

    float Y = fY;

    if (YT)
        fY = ikaros::pow(fY, 1.0/3.0);
    else
        fY = (7.787 * fY + 16/116);

    // Compute X

    float	fX = a / 500 + fY;
    bool	XT = fX > T2;
    float X;

    if (XT)
        X = ikaros::pow(fX, 3.0);
    else
        X =(fX - 16/116) / 7.787;

    // Compute Z

    float	fZ = fY - b / 200;
    bool	ZT = fZ > T2;

    float Z;
    if (ZT)
        Z = ikaros::pow(fZ, 3.0);
    else
        Z =  (fZ - 16/116) / 7.787;

    X *= 0.950456;
    Y *= 1.000000;
    Z *= 1.088754;

    red 		=   3.240479*X		- 1.537150*Y		- 0.498535*Z;
    green	= -0.969256*X		+ 1.875992*Y		+ 0.041556*Z;
    blue		=  0.055648*X		- 0.204043*Y		+ 1.057311*Z;

    red		=	scale * limit(red, 0.0, 1.0);
    green	=	scale * limit(green, 0.0, 1.0);
    blue		=	scale * limit(blue, 0.0, 1.0);
}



inline
static void
transform_rgb2lbg(float red, float green, float blue, float & L, float & b, float & g)
{
    //////
}




Module *
ColorTransform::Create(char * name, Parameter * p)
{
    return new ColorTransform(name, p);
}



ColorTransform::ColorTransform(char * name, Parameter * p):
        Module(p)
{
    AddInput("INPUT0");
    AddInput("INPUT1");
    AddInput("INPUT2");

    AddOutput("OUTPUT0");
    AddOutput("OUTPUT1");
    AddOutput("OUTPUT2");

    input0		=	NULL;
    input1		=	NULL;
    input2		=	NULL;

    output0		=	NULL;
    output1		=	NULL;
    output2		=	NULL;

    transform		=	GetIntValueFromList(p, "transform", "RGB->Lab/RGB->XYZ/Lab->RGB/XYZ->RGB");
    scale		=	GetFloatValue(p, "scale", 1.0);

    Notify(msg_verbose, "transform = %d\n", transform);
}



void
ColorTransform::SetSizes()
{
    int sx = GetInputSizeX("INPUT0");
    int sy = GetInputSizeY("INPUT0");
    if (sx != unknown_size && sy != unknown_size)
    {
        SetOutputSize("OUTPUT0", sx, sy);
        SetOutputSize("OUTPUT1", sx, sy);
        SetOutputSize("OUTPUT2", sx, sy);
    }
}



void
ColorTransform::Init()
{
    size_x	 	= GetInputSizeX("INPUT0");
    size_y	 	= GetInputSizeY("INPUT0");

    // Check equal sizes here!!!! *****

    input0		= GetInputMatrix("INPUT0");
    input1		= GetInputMatrix("INPUT1");
    input2		= GetInputMatrix("INPUT2");

    output0		= GetOutputMatrix("OUTPUT0");
    output1		= GetOutputMatrix("OUTPUT1");
    output2		= GetOutputMatrix("OUTPUT2");
}



ColorTransform::~ColorTransform()
{
}



void
ColorTransform::Tick()
{
    switch (transform)
    {
    case RGB2LAB:
        for (int i=0; i<size_x; i++)
            for (int j=0; j<size_y; j++)
                transform_rgb2lab(input0[j][i]/scale, input1[j][i]/scale, input2[j][i]/scale, output0[j][i], output1[j][i], output2[j][i]);
        break;

    case RGB2XYZ:
        for (int i=0; i<size_x; i++)
            for (int j=0; j<size_y; j++)
                transform_rgb2xyz(input0[j][i]/scale, input1[j][i]/scale, input2[j][i]/scale, output0[j][i], output1[j][i], output2[j][i]);
        break;

    case XYZ2RGB:
        for (int i=0; i<size_x; i++)
            for (int j=0; j<size_y; j++)
                transform_xyz2rgb(input0[j][i]/scale, input1[j][i]/scale, input2[j][i]/scale, output0[j][i], output1[j][i], output2[j][i]);
        break;

    case LAB2RGB:
        for (int i=0; i<size_x; i++)
            for (int j=0; j<size_y; j++)
                transform_lab2rgb(input0[j][i]/scale, input1[j][i]/scale, input2[j][i]/scale, output0[j][i], output1[j][i], output2[j][i]);
        break;
    }

    // Calculate average Lab values for the ball (after correction)
    /*
    	float rr=0, gg=0, bb=0;
    	for(int i=47; i<47+40; i++)
    		for(int j=29; j<29+40; j++)
    		{
    			rr += output0[j][i];
    			gg += output1[j][i];
    			bb += output2[j][i];
    		}

    	rr /= (40*40);
    	gg /= (40*40);
    	bb /= (40*40);

    	PrintInfo("Lab for the ball (%s):\t%.2f\t%.2f\t%.2f\n", GetName(), rr, gg, bb);
    */
}



